"""
    Unify project files structure, including:
    - custom docstring block describing file function
    - imports block
    - author block (include version, licence, copyright, maintainer and email as well)
    - code block
"""

import os
import re
import sys


author = '''
__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "0.0.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"
'''


author_ptn = re.escape(author.strip()).replace("0\.0\.0", "\d\.\d\.\d+")

py_file_struc_ptn = '(?P<service>(?:\#[^\n]+\n+)+){{,1}}\n*' \
                    '(?P<docstr>(?:\'\'\'.+?\'\'\'|""".+?""")){{,1}}\n*' \
                    '(?P<imports>(?:^(?:#|import|from|\s{{4,}})[^\n]+\n+)+){{,1}}\n*' \
                    '(?P<author>{author_ptn}){{,1}}\n*' \
                    '(?P<code>.*)'.format(author_ptn=author_ptn)
py_file_struc_re = re.compile(py_file_struc_ptn, re.M | re.S)

release_version_re = re.compile(r'\d\.\d\.\d+')

parse_paths = ['lexnlp', 'lexnlpprivate']
exclude_paths = []


def unify_file_structure(release_number):
    base_dir = os.path.normpath(os.path.join(os.path.abspath(__file__), '../..'))
    paths = [os.path.join(base_dir, p) for p in parse_paths]

    files = sorted([os.path.join(a, i)
                    for path in paths
                    for a, _, b in os.walk(path)
                    for i in b
                    if i.endswith('.py') and 'migrations' not in a])
    if exclude_paths:
        files = files + [os.path.join(base_dir, i) for i in os.listdir(base_dir)
                         if i.endswith('.py') and i not in exclude_paths]

    global author
    author = release_version_re.sub(release_number, author)

    for a_file in files:
        with open(a_file, 'r') as f:
            file_content_str = f.read()

        match = py_file_struc_re.fullmatch(file_content_str)

        if match:
            current_structure = match.groupdict()

            service = (current_structure['service'] or '').strip()
            docstr = (current_structure['docstr'] or '').strip()
            imports = (current_structure['imports'] or '').strip()
            code = (current_structure['code'] or '').strip()
        else:
            service = imports = code = docstr = ''

        new_file_content = ''.join([
            service.strip(),
            '\n'*2 if service else '',
            docstr.strip(),
            '\n'*2 if docstr else '',
            imports,
            '\n'*2 if imports else '',
            author.strip(),
            '\n'*3 if code else '',
            code,
            '\n'
        ])

        # search problems
        if new_file_content.count('__author__') > 1:
            print('>>> WARN!!! Duplicated author block: {a_file}'.format(a_file=a_file))
            # continue

        with open(a_file, 'w') as f:
            f.write(new_file_content)
        print('Done: {a_file}'.format(a_file=a_file))


if __name__ == '__main__':
    release_version = None
    args = sys.argv
    if len(args) == 1:
        print('Provide release number in format "1.2.3"')
        exit(1)
    if len(args) == 2:
        release_version = args[1]
        if not release_version_re.fullmatch(release_version):
            print('Wrong release number format "1.2.3"')
            exit(1)
    unify_file_structure(release_version)



from enum import Enum

from typing import Tuple, Generator, List

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class PhraseConstructorMethod(Enum):
    by_class = 1
    by_score = 2


class PhraseTokenClasses:
    def __init__(self,
                 outer_class: int = 0,
                 start_class: int = 1,
                 inner_class: int = 2,
                 end_class: int = 3):
        self.outer_class = outer_class
        self.start_class = start_class
        self.inner_class = inner_class
        self.end_class = end_class


class PhraseConstructorSettings:
    def __init__(self,
                 method: PhraseConstructorMethod = PhraseConstructorMethod.by_class,
                 strict: bool = False,
                 max_zeros: int = 2,
                 min_token_score: int = 2):
        self.method = method
        self.strict = strict
        self.max_zeros = max_zeros
        self.min_token_score = min_token_score

    def __repr__(self):
        if self.method == PhraseConstructorMethod.by_class:
            return f'by class, strict={self.strict}'
        return f'by score, min_score={self.min_token_score}, max_zeros={self.max_zeros}'


class PhraseConstructor:
    """
    Join "empty", "start", "middle" and "end" tokens
    into phrases.
    """

    DEFAULT_TOKEN_CLASSES = PhraseTokenClasses()

    DEFAULT_CONSTRUCTOR_SETTINGS = PhraseConstructorSettings()

    @staticmethod
    def join_tokens(tokens,
                    predicted_class,
                    feature_mask: List[int] = None,
                    settings: PhraseConstructorSettings = None,
                    token_classes: PhraseTokenClasses = None) -> Generator[Tuple[int, int], None, None]:
        settings = settings or PhraseConstructor.DEFAULT_CONSTRUCTOR_SETTINGS
        if settings.method == PhraseConstructorMethod.by_class:
            yield from PhraseConstructor.join_tokens_by_class(
                tokens, predicted_class, strict=settings.strict, token_classes=token_classes)
            return

        yield from PhraseConstructor.join_tokens_by_score(
            tokens,
            predicted_class,
            feature_mask=feature_mask,
            min_token_score=settings.min_token_score,
            max_zeros=settings.max_zeros,
            token_classes=token_classes)

    @staticmethod
    def join_tokens_by_class(
            tokens,
            predicted_class,
            strict: bool = False,
            token_classes: PhraseTokenClasses = None) -> Generator[Tuple[int, int], None, None]:
        """
        Run model on text
        """
        token_classes = token_classes or PhraseConstructor.DEFAULT_TOKEN_CLASSES
        outer_class, start_class, inner_class, end_class = (
            token_classes.outer_class, token_classes.start_class,
            token_classes.inner_class, token_classes.end_class)
        start_pos = -1

        for i in range(0, predicted_class.shape[0]):
            if predicted_class[i] == start_class and start_pos == -1:
                start_pos = i
            elif predicted_class[i] == end_class:
                if strict:
                    if start_pos >= 0:
                        yield min(tokens[start_pos][0], tokens[i][1]), max(tokens[start_pos][0], tokens[i][1])
                        start_pos = -1
                else:
                    yield min(tokens[start_pos][0], tokens[i][1]), max(tokens[start_pos][0], tokens[i][1])
                    start_pos = -1
            elif predicted_class[i] == inner_class and start_pos == -1:
                start_pos = i
            elif predicted_class[i] == outer_class and start_pos > -1:
                if not strict:
                    yield min(tokens[start_pos][0], tokens[i][1]), max(tokens[start_pos][0], tokens[i][1])
                start_pos = -1

    @staticmethod
    def join_tokens_by_score(
            tokens,
            predicted_class,
            feature_mask: List[int] = None,
            max_zeros: int = 2,
            min_token_score: int = 2,
            token_classes: PhraseTokenClasses = None) -> Generator[Tuple[int, int], None, None]:
        """
        Run model on text
        """
        token_classes = token_classes or PhraseConstructor.DEFAULT_TOKEN_CLASSES
        outer_class, start_class, inner_class, end_class = (
            token_classes.outer_class, token_classes.start_class,
            token_classes.inner_class, token_classes.end_class)

        i = -1
        while i < predicted_class.shape[0] - 1:
            i += 1
            if predicted_class[i] == start_class or predicted_class[i] == inner_class:
                token_start = i
                token_end = i
                zeros = 0
                token_score = 1 if predicted_class[i] == start_class else 0
                has_inner_tokens = False
                for j in range(i + 1, predicted_class.shape[0]):
                    if predicted_class[j] == inner_class:
                        token_end = j
                        has_inner_tokens = True
                        continue
                    if predicted_class[j] == outer_class:
                        zeros += 1
                        if zeros > max_zeros:
                            i = j
                            break
                    if predicted_class[j] == end_class:
                        token_score += 1
                        token_end = j
                        i = j
                        break
                    if predicted_class[j] == start_class:
                        i = j - 1  # continue right where stopped
                        break
                if has_inner_tokens:
                    token_score += 1

                word_start = tokens[token_start][0]
                word_end = tokens[token_end][1]
                if feature_mask:
                    for f in feature_mask[word_start: word_end]:
                        if f:
                            token_score += 1
                            break

                if token_score >= min_token_score:
                    yield tokens[token_start][0], tokens[token_end][1]



"""Transforms related to tokens for English
"""

# Imports
import collections
import os
from typing import Dict

import nltk

from lexnlp.nlp.en.tokens import get_tokens, get_stems

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


MODULE_PATH = os.path.dirname(os.path.abspath(__file__))


def get_token_distribution(text: str, lowercase=False, stopword=False) -> Dict[str, int]:
    """
    Get token distribution of text, potentially lowercasing and stopwording first.
    """
    # Iterate through tokens
    tokens = list(get_tokens(text, lowercase=lowercase, stopword=stopword))

    # Calculate distribution
    token_distribution = dict([(t, tokens.count(t)) for t in set(tokens)])
    return token_distribution


def get_stem_distribution(text: str, lowercase=False, stopword=False) -> Dict[str, int]:
    """
    Get stemmed token distribution of text, potentially lowercasing and stopwording first.
    """
    # Iterate through token stems
    tokens = list(get_stems(text, lowercase=lowercase, stopword=stopword))

    # Calculate distribution
    token_distribution = dict([(t, tokens.count(t)) for t in set(tokens)])
    return token_distribution


def get_ngram_distribution(text: str, n: int,
                           lowercase=False, stopword=False) -> Dict[str, int]:
    """
    Get n-gram distribution of text, potentially lowercasing and stopwording first.
    """

    # Return structure
    token_ngram_distribution = collections.defaultdict(int)

    # Iterate through tokens
    tokens = get_tokens(text, lowercase=lowercase, stopword=stopword)
    for ngram in nltk.ngrams(tokens, n):
        token_ngram_distribution[ngram] += 1

    # Calculate distribution
    return token_ngram_distribution


def get_bigram_distribution(text: str, lowercase=False, stopword=False) -> Dict[str, int]:
    """
    Get bigram distribution from text.
    :param text:
    :param lowercase:
    :param stopword:
    :return:
    """
    return get_ngram_distribution(text, 2, lowercase=lowercase, stopword=stopword)


def get_trigram_distribution(text: str, lowercase=False, stopword=False) -> Dict[str, int]:
    """
    Get trigram distribution from text.
    :param text:
    :param lowercase:
    :param stopword:
    :return:
    """
    return get_ngram_distribution(text, 3, lowercase=lowercase, stopword=stopword)


def get_skipgram_distribution(text: str, n: int, k: int,
                              lowercase=False, stopword=False) -> Dict[str, int]:
    """
    Get skipgram distribution from text.

    :param text:
    :param n:
    :param k:
    :param lowercase:
    :param stopword:
    :return:
    """
    # Get tokens
    tokens = get_tokens(text, lowercase=lowercase, stopword=stopword)

    # Return NLTK method results
    return nltk.util.skipgrams(tokens, n, k)



import os

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


ENV_DIRECTORY = os.path.dirname(os.path.realpath(__file__))
ENV_EN_DATA_DIRECTORY = os.path.join(ENV_DIRECTORY, 'en/data')



import regex as re
from typing import Generator, Dict, List, Any

from lexnlp.extract.common.annotations.act_annotation import ActAnnotation
from lexnlp.extract.common.annotations.text_annotation import TextAnnotation

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


ACT_PARTS_RE = re.compile(r'''
(?P<text>
    (?:sections?\s+
        (?P<section>(?:\d+(?:\(\w\))*|,\s+|,?\s+and\s+)+)\s+of\s+the\s+
    )?
    (?P<act_name>
        (?:(?:[A-Z]\w+|[A-Z&]+|and|\d+(?:[a-z]{1,3})?),?\s*)*
        (?<=\s)Act
    )
    (?:\W+|$)
    (?:of\s+(?P<year>\d{4}))?
)''', re.VERBOSE|re.MULTILINE)


def get_acts(text: str) -> Generator[Dict[str, Any], None, None]:
    for act in get_acts_annotations(text):
        yield act.to_dictionary_legacy()


def get_act_list(*args, **kwargs) -> List[Dict[str, str]]:
    return list(get_acts(*args, **kwargs))


def get_acts_annotations(text: str) -> Generator[ActAnnotation, None, None]:
    for match in ACT_PARTS_RE.finditer(text):
        captures = match.capturesdict()
        act_name = ''.join(captures.get('act_name') or [])
        year_str = ''.join(captures.get('year') or [])
        year = TextAnnotation.safe_cast(year_str, int)
        act = ActAnnotation(act_name=act_name,
                            coords=match.span(),
                            section=''.join(captures.get('section') or []),
                            year=year,
                            ambiguous=act_name == 'Act',
                            text=''.join(captures.get('text') or []),
                            locale='en')
        yield act


def get_acts_annotations_list(text: str) -> List[ActAnnotation]:
    return list(get_acts_annotations(text))



# pylint: disable=unused-import

from typing import List, Generator, Tuple
import regex as re
from lexnlp.extract.common.copyrights.copyright_en_style_parser import CopyrightEnStyleParser
from lexnlp.extract.common.annotations.copyright_annotation import CopyrightAnnotation
from lexnlp.extract.de.language_tokens import DeLanguageTokens
from lexnlp.utils.lines_processing.line_processor import LineSplitParams, LineProcessor

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class CopyrightDeParser(CopyrightEnStyleParser):
    line_processor = None  # LineProcessor
    copyright_words = ['Copyright', 'Urheberrechte', 'Urheberschutz', 'Eigentumsrecht']
    copyright_words_ptrn = r'\W\s*|'.join(copyright_words)

    copyright_ptn = fr"(({copyright_words_ptrn}|\(\s*[Cc]\s*\)\s*|©)+\s*{CopyrightEnStyleParser.year_ptn}?\s*(.+))"
    copyright_ptn_re = re.compile(copyright_ptn)

    @staticmethod
    def init_parser():
        split_params = LineSplitParams()
        split_params.line_breaks = {'\n', '.', ';', '!', '?'}
        split_params.abbreviations = DeLanguageTokens.abbreviations
        split_params.abbr_ignore_case = True
        CopyrightDeParser.line_processor = LineProcessor(line_split_params=split_params)

    @classmethod
    def extract_phrases_with_coords(cls, sentence: str) -> List[Tuple[str, int, int]]:
        return [(t.text, t.start, t.get_end()) for t in
                cls.line_processor.split_text_on_line_with_endings(sentence)]


CopyrightDeParser.init_parser()


def get_copyright_annotations(text: str, return_sources=False) -> \
        Generator[CopyrightAnnotation, None, None]:
    for ant in  CopyrightDeParser.get_copyright_annotations(text,
                                                            return_sources):
        ant.locale = 'de'
        yield ant


#def get_copyright_annotations(text: str, language: str = None) -> Generator[dict, None, None]:
#    ants = parser.parse(text, language if language else 'de')
#    for ant in ants:
#        yield ant


def get_copyrights(text: str, return_sources=False) -> \
        Generator[dict, None, None]:
    for ant in get_copyright_annotations(text, return_sources):
        yield ant.to_dictionary()


def get_copyright_list(text: str, return_sources=False) -> List[CopyrightAnnotation]:
    return list(get_copyright_annotations(text, return_sources))



"""Title segmentation for English.

This module implements title segmentation/location in English using simple
machine learning classifiers.
"""

# Imports
import os
import string
from typing import Generator

# Packages
import pandas
import sklearn.ensemble
from sklearn.externals import joblib

# Project
from lexnlp.nlp.en.segments.utils import build_document_line_distribution
from lexnlp.utils.decorators import safe_failure
from lexnlp.utils.unicode.unicode_lookup import UNICODE_CHAR_TOP_CATEGORY_MAPPING

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


# Setup module path
MODULE_PATH = os.path.dirname(os.path.abspath(__file__))

# Load segmenters
SECTION_SEGMENTER_MODEL = joblib.load(os.path.join(MODULE_PATH, "./title_locator.pickle"))


def build_title_features(lines, line_id, line_window_pre, line_window_post, characters=string.printable,
                         include_doc=None):
    """
    Build a feature vector for a given line ID with given parameters.

    :param lines:
    :param line_id:
    :param line_window_pre:
    :param line_window_post:
    :param characters:
    :param include_doc:
    :return:
    """
    # Feature vector
    feature_vector = {}

    # Check start offset
    if line_id < line_window_pre:
        line_window_pre = line_id

    # Check final offset
    if (line_id + line_window_post) >= len(lines):
        line_window_post = len(lines) - line_window_post - 1

    # Iterate through window
    for i in range(0 - line_window_pre, line_window_post + 1):
        try:
            line = lines[line_id + i]
        except IndexError:
            continue

        index_str = str(i)
        # Count length
        feature_vector["line_len_" + index_str] = len(line)
        feature_vector["line_lenstrip_" + index_str] = len(line.strip())
        feature_vector["line_title_case_" + index_str] = line == line.title()
        feature_vector["line_upper_case_" + index_str] = line.isupper()

        alpha_count, number_count, punct_count, whitespace_count = 0, 0, 0, 0
        for c in line:
            if UNICODE_CHAR_TOP_CATEGORY_MAPPING[c] == 'L':
                alpha_count += 1
            elif UNICODE_CHAR_TOP_CATEGORY_MAPPING[c] == 'Z':
                whitespace_count += 1
            elif UNICODE_CHAR_TOP_CATEGORY_MAPPING[c] == 'N':
                number_count += 1
            elif UNICODE_CHAR_TOP_CATEGORY_MAPPING[c] == 'P':
                punct_count += 1

        # Count characters
        feature_vector["line_n_alpha_" + index_str] = alpha_count
        feature_vector["line_n_number_" + index_str] = number_count
        feature_vector["line_n_punct_" + index_str] = punct_count
        feature_vector["line_n_whitespace_" + index_str] = whitespace_count

    # Simple checks
    line = lines[line_id]
    line_strip_lower = line.strip().lower()
    feature_vector["agreement"] = 1 if "agreement" in line else 0
    feature_vector["Agreement"] = 1 if "Agreement" in line else 0
    feature_vector["AGREEMENT"] = 1 if "AGREEMENT" in line else 0
    feature_vector["contract"] = 1 if "contract" in line else 0
    feature_vector["Contract"] = 1 if "Contract" in line else 0
    feature_vector["CONTRACT"] = 1 if "CONTRACT" in line else 0
    feature_vector["amendment"] = 1 if "amendment" in line else 0
    feature_vector["Amendment"] = 1 if "Amendment" in line else 0
    feature_vector["AMENDMENT"] = 1 if "AMENDMENT" in line else 0
    feature_vector["ew_agreement"] = 1 if line_strip_lower.endswith("agreement") else 0
    feature_vector["sw_amendment"] = 1 if line_strip_lower.startswith("amendment") else 0

    # Build character vector
    characters_count = {}
    for character in characters:
        characters_count[character] = 0

    for character in lines[line_id]:
        if characters_count.get(character) is not None:
            characters_count[character] += 1

    for character, count in characters_count.items():
        feature_vector["char_" + character] = count

    # Add doc if requested
    if include_doc:
        feature_vector.update(include_doc)

    return feature_vector


def build_document_title_features(text, window_pre=3, window_post=3):
    """
    Get a document title given file text.
    """
    # Get document character distribution
    doc_distribution = build_document_line_distribution(text)

    # Parse all lines
    lines = text.splitlines()
    feature_data = []

    for line_id in range(len(lines)):
        feature_data.append(build_title_features(lines, line_id, window_pre, window_post, include_doc=doc_distribution))

    # Get feature DF
    feature_df = pandas.DataFrame(feature_data).fillna(-1).astype(int)
    return feature_df


def build_model(training_file_path):
    """
    Build a title extraction model given a training file path.

    :param training_file_path:
    :return:
    """
    import requests
    import numpy

    # Read title training data
    training_data = pandas.read_csv(training_file_path, encoding="utf-8", low_memory=False)
    training_data = training_data.loc[-training_data["Line Number"].isnull(), :]
    training_data.head()

    # All data
    all_feature_list = []
    all_target_list = []
    all_file_lines = []

    # Build training data
    for _, row in training_data.iterrows():
        # Download file
        file_url = row["File"].replace("https://github.com/", "https://raw.githubusercontent.com/").replace("/blob/",
                                                                                                            "/")
        file_text = requests.get(file_url).text
        file_lines = file_text.splitlines()

        # Get features and target for model
        feature_data = build_document_title_features(file_text)
        target_data = pandas.Series(numpy.zeros((feature_data.shape[0],)))

        # Parse line numbers
        if "-" in row["Line Number"]:
            target_line_ranges = row["Line Number"].split("-")
            target_lines = list(range(int(target_line_ranges[0]),
                                      int(target_line_ranges[1]) + 1))
        else:
            target_lines = [int(row["Line Number"])]

        if len(target_lines) > 2:
            continue

        for target_line_num in target_lines:
            if len(file_lines[target_line_num - 1].strip()) > 0:
                target_data.iloc[target_line_num - 1] = 1

        # Append
        all_feature_list.append(feature_data)
        all_target_list.append(target_data)
        all_file_lines.extend((row["File"], l) for l in file_lines)

    # Collate
    all_feature_df = pandas.concat(all_feature_list, axis=0)
    all_target_df = pandas.concat(all_target_list, axis=0)

    # Build final model
    model = sklearn.ensemble.ExtraTreesClassifier(n_estimators=25)
    model.fit(all_feature_df, all_target_df)

    # Save production model
    joblib.dump(model, "title_locator.pickle")


@safe_failure
def get_titles(text, window_pre=3, window_post=3, score_threshold=0.5) -> Generator:
    """
    Get titles from text.
    :param text:
    :param window_pre:
    :param window_post:
    :param score_threshold:
    :return:
    """

    # Get features and target for model
    feature_data = build_document_title_features(text, window_pre, window_post)

    # Predict title lines
    predicted_lines = SECTION_SEGMENTER_MODEL.predict_proba(feature_data)
    predicted_df = pandas.DataFrame(predicted_lines, columns=["prob_false", "prob_true"])
    title_lines = predicted_df.loc[predicted_df["prob_true"] >= score_threshold, :].index.tolist()

    # Check if results
    if len(title_lines) > 0:
        # Get lines
        lines = text.splitlines()

        # Iterate through lines
        title = ""
        for i in range(len(lines)):
            if i in title_lines:
                title += lines[i] + " "
            elif len(lines[i].strip()) == 0:
                continue
            else:
                if len(title) > 0:
                    yield title.strip()
                    title = ""

        if len(title) > 0:
            yield title.strip()



# pylint: disable=unused-argument

import os
import re
from typing import Generator, List
from lexnlp.extract.common.annotations.court_annotation import CourtAnnotation
from lexnlp.extract.common.universal_court_parser import UniversalCourtsParser, ParserInitParams
from lexnlp.extract.de.language_tokens import DeLanguageTokens
from lexnlp.utils.lines_processing.line_processor import LineSplitParams

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


def setup_de_parser():
    def preproc_func(text):
        return re.sub('e$', '[e]?', text)

    ptrs = ParserInitParams()
    ptrs.key_word_preproc_func = preproc_func
    ptrs.court_pattern_checker = re.compile('gericht')
    ptrs.split_ptrs = LineSplitParams()
    ptrs.split_ptrs.line_breaks = {'\n', '.', ';', ','}.union(set(DeLanguageTokens.conjunctions))
    ptrs.split_ptrs.abbreviations = DeLanguageTokens.abbreviations
    ptrs.split_ptrs.abbr_ignore_case = True

    ptrs.column_names = {'type': 'Court Type (de-DE)',
                         'name': 'Court Name (de-DE)',
                         'jurisdiction': 'Jurisdiction',
                         'alias': 'Alias (de-DE)'}

    path = os.path.join(os.path.dirname(__file__), "../../config/de/de_courts.csv")
    ptrs.dataframe_paths = [path]
    return UniversalCourtsParser(ptrs)


parser = setup_de_parser()


def get_court_annotations(text: str, language: str = None) -> \
        Generator[CourtAnnotation, None, None]:
    yield from parser.parse(text, language if language else 'de')


def get_courts(text: str, language=None) -> Generator[dict, None, None]:
    courts = parser.parse(text, language if language else 'de')
    for c in courts:
        yield c.to_dictionary()


def get_court_list(text: str, language=None) -> List[CourtAnnotation]:
    return parser.parse(text, language if language else 'de')



"""Entity extraction for English using NLTK and NLTK pre-trained maximum entropy classifier.

This module implements basic entity extraction functionality in English relying on the pre-trained
NLTK functionality, including POS tagger and NE (fuzzy) chunkers.

Todo:
  * Better define interface for sentences vs. raw text
  * Standardize generator vs list
"""

# pylint: disable=W0612

# Imports
import re
import string

from typing import Generator, Dict, Tuple

import nltk

from lexnlp.extract.en.entities.nltk_tokenizer import NltkTokenizer
from lexnlp.extract.common.annotations.phrase_position_finder import PhrasePositionFinder
from lexnlp.extract.common.annotations.company_annotation import CompanyAnnotation
from lexnlp.config.en.company_types import COMPANY_TYPES, COMPANY_DESCRIPTIONS
from lexnlp.extract.en.entities import nltk_re
from lexnlp.extract.en.utils import strip_unicode_punctuation, NPExtractor
from lexnlp.nlp.en.segments.sentences import get_sentence_list, get_sentence_span_list
from lexnlp.nlp.en.tokens import get_token_list

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


VALID_PUNCTUATION = [",", ".", "&"]

COMPANY_TYPES = sorted(list(COMPANY_TYPES.keys()) + COMPANY_DESCRIPTIONS, key=len)
COMPANY_TYPES_RE = re.compile(r' %s(?:\W|$)' % '|'.join([re.escape(i.strip('.'))
                                                         for i in COMPANY_TYPES]), re.IGNORECASE)

PERSONS_STOP_WORDS = re.compile(
    r'avenue|amendment|agreement|addendum|article|assignment|exhibit', re.IGNORECASE)


def contains_companies(person:str, companies) -> bool:
    if COMPANY_TYPES_RE.search(person):
        # noinspection PyTypeChecker
        for ant in nltk_re.get_companies(person):  # type: CompanyAnnotation
            if ant.name == ant.company_type or ant.name == ant.description:
                continue
            return True

    for ant in companies:
        # Solving this scenario: This Amendment to Employment Agreement ("Amendment") is entered into
        # between Marsh Supermarkets, Inc. (the "Company"), and Don E. Marsh (the "Executive").
        # because that is pretty common , even though it screws up this scenario
        # "This is an agreement between John Smith and John Smith, LLC"
        if person in ant.name:
            return True
    return False


def get_persons(text, strict=False, return_source=False, window=2) -> Generator:
    """
    Get names from text.
    :param window:
    :param return_source:
    :param strict:
    :param text:
    :return:
    """
    # Iterate through sentences
    for sentence in get_sentence_list(text):
        # Tag sentence
        sentence_pos = nltk.pos_tag(get_token_list(sentence))
        companies = list(get_company_annotations(text))

        # Iterate through chunks
        persons = []
        last_person_pos = None

        for i, chunk in enumerate(nltk.ne_chunk(sentence_pos)):
            if type(chunk) == nltk.tree.Tree:
                # Check label
                if chunk.label() == 'PERSON':
                    if not strict and last_person_pos is not None and (i - last_person_pos) < window:
                        persons[-1] += " " + " ".join([c[0] for c in chunk])
                    else:
                        persons.append(" ".join([c[0] for c in chunk]))
                    last_person_pos = i
            elif not strict and last_person_pos is not None and (i - last_person_pos) < window:
                if chunk[1] in ["NNP", "NNPS"]:
                    persons[-1] += " " + chunk[0]
                    last_person_pos = i
                elif chunk[1] in ["CC"] or chunk[0] in VALID_PUNCTUATION:
                    if chunk[0].lower() in ["or"]:
                        continue
                    persons[-1] += (" " if chunk[0].lower() in ["&", "and"] else "") + chunk[0]
                    last_person_pos = i
                else:
                    last_person_pos = None

        # Cleanup and yield
        for person in persons:
            # Cleanup
            person = person.strip()
            if len(person) <= 2:
                continue

            if PERSONS_STOP_WORDS.search(person):
                continue

            person = strip_unicode_punctuation(person).strip(string.punctuation).strip(string.whitespace)

            if contains_companies(person, companies):
                continue

            if person.lower().endswith(" and"):
                person = person[0:-4]
            elif person.endswith(" &"):
                person = person[0:-2]


            if return_source:
                yield person, sentence
            else:
                yield person


def get_geopolitical(text, strict=False, return_source=False, window=2) -> Generator:
    """
    Get GPEs from text.
    :param window:
    :param return_source:
    :param strict:
    :param text:
    :return:
    """
    # Iterate through sentences
    for sentence in get_sentence_list(text):
        # Tag sentence
        sentence_pos = nltk.pos_tag(get_token_list(sentence))

        # Iterate through chunks
        gpes = []
        last_gpe_pos = None
        for i, chunk in enumerate(nltk.ne_chunk(sentence_pos)):
            if type(chunk) == nltk.tree.Tree:
                # Check label
                if chunk.label() == 'GPE':
                    if not strict and last_gpe_pos is not None and (i - last_gpe_pos) < window:
                        gpes[-1] += " " + " ".join([c[0] for c in chunk])
                    else:
                        gpes.append(" ".join([c[0] for c in chunk]))
                    last_gpe_pos = i
            elif not strict and last_gpe_pos is not None and (i - last_gpe_pos) < window:
                if chunk[1] in ["NNP", "NNPS"]:
                    gpes[-1] += " " + chunk[0]
                    last_gpe_pos = i
                elif chunk[1] in ["CC"] or chunk[0] in VALID_PUNCTUATION:
                    if chunk[0].lower() in ["or"]:
                        continue
                    gpes[-1] += (" " if chunk[0].lower() in ["&", "and"] else "") + chunk[0]
                    last_gpe_pos = i
                else:
                    last_gpe_pos = None

        # Clean up names and yield
        for gpe in gpes:
            # Cleanup
            gpe = gpe.strip()
            if len(gpe) <= 2:
                continue

            if gpe.lower().endswith(" and"):
                gpe = gpe[0:-4]
            elif gpe.endswith(" &"):
                gpe = gpe[0:-2]

            gpe = strip_unicode_punctuation(gpe).strip(string.punctuation).strip(string.whitespace)
            if return_source:
                yield gpe, sentence
            else:
                yield gpe


def get_noun_phrases(text, strict=False, return_source=False, window=3, valid_punctuation=None) -> Generator:
    """
    Get NNP phrases from text.
    :param window:
    :param return_source:
    :param strict:
    :param text:
    :return:
    """
    valid_punctuation = valid_punctuation or VALID_PUNCTUATION
    # Iterate through sentences
    for sentence in get_sentence_list(text):
        # Tag sentence
        sentence_pos = nltk.pos_tag(get_token_list(sentence))

        # Iterate through chunks
        nnps = []
        last_nnp_pos = None
        for i, chunk in enumerate(sentence_pos):
            do_join = not strict and last_nnp_pos is not None and (i - last_nnp_pos) < window
            # Check label
            if chunk[1] in ["NNP", "NNPS"]:
                if do_join:
                    sep = "" if "(" in valid_punctuation and nnps[-1][-1] == "(" else " "
                    nnps[-1] += sep + chunk[0]
                else:
                    nnps.append(chunk[0])
                last_nnp_pos = i
            elif do_join:
                if chunk[1] in ["CC"] or chunk[0] in valid_punctuation:
                    if chunk[0].lower() in ["or"]:
                        continue
                    nnps[-1] += (" " if chunk[0].lower() in ["&", "and", "("] else "") + chunk[0]
                    last_nnp_pos = i
                else:
                    last_nnp_pos = None

        # Clean up names and yield
        for nnp in nnps:
            # Cleanup
            nnp = nnp.strip()
            if len(nnp) <= 2:
                continue

            if nnp.lower().endswith(" and"):
                nnp = nnp[0:-4].strip()
            elif nnp.endswith(" &"):
                nnp = nnp[0:-2].strip()

            nnp = strip_unicode_punctuation(nnp).strip(string.punctuation).strip(string.whitespace)
            if return_source:
                yield nnp, sentence
            else:
                yield nnp


class CompanyNPExtractor(NPExtractor):

    def __init__(self, grammar=None):
        grammar = grammar or r"""
            NBAR:
                {<DT>?<NNP.*|JJ|POS|\(|\)|,>*<NNP.*>}  # DeTerminer, Proper Noun, Adjective, brackets, terminated by Proper Noun
            IN:
                {<CC|IN>}   # Coordinating Conjunction, Preposition/Subordinating Conjunction
            NP:
                {(<NBAR><IN>)*<NBAR><VBD>*}  # Delaver Housing Incorporated (NNP-NNP-VBD)
        """
        super().__init__(grammar)

    def get_tokenizer(self):
        orig_tokenizer = nltk.tokenize.TreebankWordTokenizer
        punctuation = list(orig_tokenizer.PUNCTUATION)
        punctuation[4] = (re.compile(r'[;@#$%]', re.UNICODE), ' \\g<0> ')
        # for case like "McDonald\'s Incorporated: Burgers" when POS tokenizer treats
        # "Incorporated:" as VBN instead of NNP and NP extractor fails to recognize such grammar
        punctuation.append((re.compile(r':'), r';'))
        # for case when apostrophe is in company name like Moody`s
        starting_quotes = [(re.compile(r'`s '), r'-ES-')] + list(orig_tokenizer.STARTING_QUOTES) + [
            (re.compile(r'-ES-'), r'`s ')]

        tokenizer = NltkTokenizer(punctuation, starting_quotes)
        return tokenizer.tokenize

    @staticmethod
    def strip_np(np):
        return np.strip(string.punctuation + string.whitespace)

    def cleanup_leaves(self, leaves):
        leaves = super().cleanup_leaves(leaves)
        leaves = [i for i in leaves if i[0] != ('a', 'DT') and i[0] != ('A', 'DT') and
                  not (i[0][1] == 'JJ' and i[0][0].islower())]
        return leaves


np_extractor = CompanyNPExtractor()


def get_company_annotations(
        text: str,
        strict: bool = False,
        use_gnp: bool = False,
        count_unique: bool = False,
        name_upper: bool = False,
        ) -> Generator[CompanyAnnotation, None, None]:
    """
    Find company names in text, optionally using the stricter article/prefix expression.
    :param parse_name_abbr:
    :param text:
    :param strict:
    :param use_gnp: use get_noun_phrases or NPExtractor
    :param name_upper: return company name in upper case.
    :param count_unique: return only unique companies - case insensitive.
    :return:
    """
    # skip if all text is in uppercase
    if text == text.upper():
        return
    valid_punctuation = VALID_PUNCTUATION + ["(", ")"]

    unique_companies = {}  # type: Dict[Tuple[str, str], CompanyAnnotation]

    if COMPANY_TYPES_RE.search(text):
        # Iterate through sentences
        for s_start, s_end, sentence in get_sentence_span_list(text):
            # skip if whole phrase is in uppercase
            if sentence == sentence.upper():
                continue
            if use_gnp:
                phrases = list(get_noun_phrases(sentence, strict=strict,
                                                valid_punctuation=valid_punctuation))
            else:
                phrases = list(np_extractor.get_np(sentence))
            phrase_spans = PhrasePositionFinder.find_phrase_in_source_text(sentence, phrases)

            for phrase, p_start, p_end in phrase_spans:
                if COMPANY_TYPES_RE.search(phrase):
                    # noinspection PyTypeChecker
                    for ant in nltk_re.get_companies(
                            phrase, use_sentence_splitter=False):  # type: CompanyAnnotation

                        if ant.name == ant.company_type or ant.name == ant.description:
                            continue
                        ant.coords = (ant.coords[0] + s_start + p_start,
                                      ant.coords[1] + s_start + p_start)

                        if name_upper:
                            ant.name = ant.name.upper()

                        if count_unique:
                            unique_key = (ant.name.lower() if ant.name else None, ant.company_type_abbr)
                            existing_result = unique_companies.get(unique_key)

                            if existing_result:
                                existing_result.counter += 1
                            else:
                                unique_companies[unique_key] = ant
                        else:
                            yield ant

        if count_unique:
            for company in unique_companies.values():
                yield company

    # search for acronyms in text ("[A-Z]" or (A-Z))
    # try to merge annotations (in one sentence!) in acronyms


def get_companies(text: str,
                  strict: bool = False,
                  use_gnp: bool = False,
                  detail_type: bool = False,
                  count_unique: bool = False,
                  name_upper: bool = False,
                  parse_name_abbr: bool = False,
                  return_source: bool = False):
    """
    Find company names in text, optionally using the stricter article/prefix expression.
    :param text:
    :param strict:
    :param use_gnp: use get_noun_phrases or NPExtractor
    :param detail_type: return detailed type (type, unified type, label) vs type only
    :param name_upper: return company name in upper case.
    :param count_unique: return only unique companies - case insensitive.
    :param parse_name_abbr: return company abbreviated name if exists.
    :param return_source:
    :return:
    """
    # skip if all text is in uppercase
    # noinspection PyTypeChecker
    for ant in get_company_annotations(text,
                                       strict,
                                       use_gnp,
                                       count_unique,
                                       name_upper):  # type:CompanyAnnotation
        result = (ant.name, ant.company_type)
        if detail_type:
            result += (ant.company_type_abbr, ant.company_type_label, ant.description)
        if parse_name_abbr:
            result += (ant.name_abbr,)
        if return_source and not count_unique:
            result = result + (ant.text,)
        if count_unique:
            result = result + (ant.counter,)
        yield result



import codecs

from typing import Set

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class LanguageDictionaryReader:
    """
    This class reads text files, where values are separated by <line_breaks>,
    strips the values if needed and returns them as List or Dict.

    We use this class, e.g., while reading De locale common abbreviations.
    """

    @staticmethod
    def read_str_set(file_path: str,
                     encoding='utf8',
                     strip_symbols=' ') -> Set[str]:
        words = set()
        with codecs.open(file_path, encoding=encoding, mode='r') as fr:
            for line in fr.readlines():
                if not line:
                    continue
                word = line.strip('\n')
                if not word:
                    continue
                if strip_symbols:
                    word = word.strip(strip_symbols)
                words.add(word)

        return words



"""Section segmentation for English.

This module implements section segmentation in English using simple
machine learning classifiers.

Todo:
  * Standardize model (re-)generation
"""

# Imports
import os
import string
import unicodedata

from typing import Generator

# Packages
import pandas
import regex as re
from sklearn.externals import joblib

# Project imports
from lexnlp.nlp.en.segments.utils import build_document_line_distribution
from lexnlp.utils.map import Map
from lexnlp.utils.decorators import safe_failure

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


# Setup module path
MODULE_PATH = os.path.dirname(os.path.abspath(__file__))

# Load segmenters
SECTION_SEGMENTER_MODEL = joblib.load(os.path.join(MODULE_PATH, "./section_segmenter.pickle"))


def build_section_break_features(lines, line_id, line_window_pre, line_window_post, characters=string.printable,
                                 include_doc=None):
    """
    Build a feature vector for a given line ID with given parameters.

    :param lines:
    :param line_id:
    :param line_window_pre:
    :param line_window_post:
    :param characters:
    :param include_doc:
    :return:
    """
    # Feature vector
    feature_vector = {}

    # Check start offset
    if line_id < line_window_pre:
        line_window_pre = line_id

    # Check final offset
    if (line_id + line_window_post) >= len(lines):
        line_window_post = len(lines) - line_window_post - 1

    # Iterate through window
    for i in range(-line_window_pre, line_window_post + 1):
        try:
            line = lines[line_id + i]
        except IndexError:
            continue

        # Count length
        feature_vector["line_len_{0}".format(i)] = len(line)
        feature_vector["line_lenstrip_{0}".format(i)] = len(line.strip())
        feature_vector["line_title_case_{0}".format(i)] = line == line.title()
        feature_vector["line_upper_case_{0}".format(i)] = line == line.upper()

        # Count characters
        feature_vector["line_n_alpha_{0}".format(i)] = sum([1 for c in line if unicodedata.category(c).startswith("L")])
        feature_vector["line_n_number_{0}".format(i)] = sum(
            [1 for c in line if unicodedata.category(c).startswith("N")])
        feature_vector["line_n_punct_{0}".format(i)] = sum([1 for c in line if unicodedata.category(c).startswith("P")])
        feature_vector["line_n_whitespace_{0}".format(i)] = sum(
            [1 for c in line if unicodedata.category(c).startswith("Z")])

    # Simple checks
    line = lines[line_id]
    feature_vector["section"] = 1 if "section" in line else 0
    feature_vector["SECTION"] = 1 if "SECTION" in line else 0
    feature_vector["Section"] = 1 if "Section" in line else 0
    feature_vector["article"] = 1 if "article" in line else 0
    feature_vector["ARTICLE"] = 1 if "ARTICLE" in line else 0
    feature_vector["Article"] = 1 if "Article" in line else 0
    feature_vector["sw_section"] = 1 if line.strip().lower().startswith("section") else 0
    feature_vector["sw_article"] = 1 if line.strip().lower().startswith("article") else 0
    feature_vector["first_char_punct"] = (line.strip()[0] in string.punctuation) if len(line.strip()) > 0 else False
    feature_vector["last_char_punct"] = (line.strip()[-1] in string.punctuation) if len(line.strip()) > 0 else False
    feature_vector["first_char_number"] = (line.strip()[0] in string.digits) if len(line.strip()) > 0 else False
    feature_vector["last_char_number"] = (line.strip()[-1] in string.digits) if len(line.strip()) > 0 else False

    # Build character vector
    for character in characters:
        feature_vector["char_{0}".format(character)] = lines[line_id].count(character)

    # Add doc if requested
    if include_doc:
        feature_vector.update(include_doc)

    return feature_vector


@safe_failure
def get_sections(text, window_pre=3, window_post=3, score_threshold=0.5) -> Generator:
    """
    Get sections from text.
    NLP-based detection of sections.
    :param text:
    :param window_pre:
    :param window_post:
    :param score_threshold:
    :return:
    """

    # Get document character distribution
    doc_distribution = build_document_line_distribution(text)
    lines = text.splitlines()
    test_feature_data = []
    for line_id in range(len(lines)):
        test_feature_data.append(
            build_section_break_features(lines, line_id, window_pre, window_post, include_doc=doc_distribution))

    # Predict page breaks
    test_feature_df = pandas.DataFrame(test_feature_data).fillna(-1)
    test_predicted_lines = SECTION_SEGMENTER_MODEL.predict_proba(test_feature_df)
    predicted_df = pandas.DataFrame(test_predicted_lines, columns=["prob_false", "prob_true"])
    section_breaks = predicted_df.loc[predicted_df["prob_true"] >= score_threshold, :].index.tolist()

    if len(section_breaks) > 0:
        # Get first break
        pos0 = 0
        pos1 = section_breaks[0]
        section = "\n".join(lines[pos0:pos1])
        if len(section.strip()) > 0:
            yield section

        # Iterate through section breaks
        for i in range(len(section_breaks) - 1):
            # Get breaks
            pos0 = section_breaks[i]
            pos1 = section_breaks[i + 1]
            # Get text
            section = "\n".join(lines[pos0:pos1])
            if len(section.strip()) > 0:
                yield section

        # Yield final section
        section = "\n".join(lines[section_breaks[-1]:])
        if len(section.strip()) > 0:
            yield section


SECTION_TITLE_PTN = r"""
\s*
(
    (?i:(?:appendix|part|section|article|(?:sub)?title|EXHIBIT|SCHEDULE)\s+)?
    (?-i:\d+(?:\.\d+)?
         |
         \p{Lu}(?:-\d+(?:\.\d+)?)?
         |
         [XVILCM]+
         |
         \((?:\d{1,3}|\p{L}+)\)
    )
)
(?:\.|\s|$)
"""
SECTION_TITLE_RE1 = re.compile(r'(?<=[,\.:>;\d\n\s]\n)' + SECTION_TITLE_PTN, re.M | re.X)
SECTION_TITLE_RE2 = re.compile(r'\A' + SECTION_TITLE_PTN, re.M | re.X)


@safe_failure
def get_sections_re(text) -> Generator:
    """
    Get sections from text.
    Regex-based detection of text sections.
    :param text: str - source full text
    :return: generator of str
    """
    prev_start = None
    for match in SECTION_TITLE_RE1.finditer(text):
        start = match.start()
        if prev_start:
            yield text[prev_start:start]
        elif start != 0:
            yield text[0:start]
        prev_start = start
    if prev_start:
        yield text[prev_start:]


@safe_failure
def get_section_spans(text, use_ml=True,
                      return_text=True, skip_empty_headers=False,
                      sections_hierarchy=None) -> Generator:
    """
    Get sections from text.
    Use NLP-based detection OR regex-bases detection of sections - see use_ml param.
    :param text: str - source full text
    :param use_ml: bool - use sklearn classifier otherwise use regex-based detection
    :param return_text: bool - return section text
    :param skip_empty_headers: bool - return results containing headers only
    :param sections_hierarchy: list of regexes
    :return: Generator of dicts
    """

    _start_index_counter = 0
    level_parser = SectionLevelParser(sections_hierarchy=sections_hierarchy)
    sections_detector = get_sections if use_ml else get_sections_re

    for section in sections_detector(text):
        start_index = _start_index_counter + text[_start_index_counter:].index(section)
        end_index = start_index + len(section)
        _start_index_counter = end_index
        try:
            title = SECTION_TITLE_RE2.findall(section)[0]
            title_start = start_index + section.index(title)
            title_end = title_start + len(title)
        except IndexError:
            # wrong number of features (short text or smth similar)
            title = title_start = title_end = None
        if skip_empty_headers and not title:
            continue

        # try to get level
        level, abs_level = level_parser.detect(title)

        res = dict(
            start=start_index,
            end=end_index,
            title=title,
            title_start=title_start,
            title_end=title_end,
            level=level,
            abs_level=abs_level,
        )
        if return_text:
            res['text'] = section
        yield res


class SectionLevelParser:

    DEFAULT_SECTION_HIERARCHY = [
        r'(?i:(appendix|exhibit|schedule|part|title)\s+\S+)',
        r'(?i:subtitle\s+\S+)',
        r'(?i:section\s+\S+)',
        r'(?i:subsection\s+\S+)',
        r'(?i:article\s+\S+)',
        r'\p{Lu}+(?:-\d+(?:\.\d+)?)?',
        r'[\d\.]+',
        r'\p{L}+(?:-\d+(?:\.\d+)?)?',
        r'\([\p{L}\d]+\)'
    ]

    def __init__(self, sections_hierarchy=None):
        if not sections_hierarchy:
            sections_hierarchy = self.DEFAULT_SECTION_HIERARCHY
        self.default_sections_hierarchy = [Map(regex=re.compile(i),
                                               abs_level=n,
                                               rel_level=None)
                                           for n, i in enumerate(sections_hierarchy, start=1)]
        self.level = 1        # represents previous->current relative level (new custom hierarchy)
        self.abs_level = 1    # represents previous->current absolute level from sections_hierarchy
        self.prev_level_re = None
        self.title = ''

    @property
    def current_sections_hierarchy(self):
        return [i for i in self.default_sections_hierarchy if i.rel_level]

    def detect(self, title):
        self.title = title
        if not title:
            return 0, 0
        elif self.prev_level_re and self.prev_level_re.match(title):
            return self.level, self.abs_level
        elif self.current_sections_hierarchy:
            levels = self.get_from_detected()
            if levels is not None:
                return levels
        self.get_from_default()
        return self.level, self.abs_level

    def get_from_detected(self):
        for level_data in self.current_sections_hierarchy:
            if level_data.regex.match(self.title):
                self.level = level_data.rel_level
                self.abs_level = level_data.abs_level
                self.prev_level_re = level_data.regex
                return self.level, self.abs_level

    def get_from_default(self):
        for level_data in self.default_sections_hierarchy:
            if level_data.regex.match(self.title):
                if level_data.abs_level > self.abs_level:
                    self.level += 1
                elif level_data.abs_level < self.abs_level:
                    self.level -= 1
                level_data.rel_level = self.level
                self.abs_level = level_data.abs_level
                self.prev_level_re = level_data.regex
                break



"""Token parsing for English.

This module implements token parsing, such as tokens, stems, and lemma tokenization functionality in English.

Todo:
"""

# Imports
import os
import pickle
from typing import List, Generator

# NLTK imports
import nltk
from nltk.corpus import wordnet

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


MODULE_PATH = os.path.dirname(os.path.abspath(__file__))

# Stopwords
STOPWORDS = pickle.load(open(os.path.join(MODULE_PATH, "stopwords.pickle"), "rb"))

# Collocations
COLLOCATION_SIZE = 10000
BIGRAM_COLLOCATIONS = pickle.load(
    open(os.path.join(MODULE_PATH, "collocation_bigrams_{0}.pickle".format(COLLOCATION_SIZE)), "rb"))
TRIGRAM_COLLOCATIONS = pickle.load(
    open(os.path.join(MODULE_PATH, "collocation_trigrams_{0}.pickle".format(COLLOCATION_SIZE)), "rb"))

# Setup default stemmer for English
DEFAULT_STEMMER = nltk.stem.snowball.EnglishStemmer()

# Setup lemmatizers for English
DEFAULT_LEMMATIZER = nltk.stem.wordnet.WordNetLemmatizer()


def get_wordnet_pos(treebank_tag):
    """
    Return wordnet POS object from Treebank POS tag.
    :param treebank_tag:
    :return:
    """
    if treebank_tag.startswith('J'):
        return wordnet.ADJ
    elif treebank_tag.startswith('V'):
        return wordnet.VERB
    elif treebank_tag.startswith('N'):
        return wordnet.NOUN
    elif treebank_tag.startswith('R'):
        return wordnet.ADV
    else:
        return None


def get_tokens(text, lowercase=False, stopword=False, preserve_line=True) -> Generator:
    """
    Get token generator from text.
    :param text:
    :param lowercase:
    :param stopword:
    :param preserve_line: keep the preserve the sentence and not sentence tokenize it.
    :return:
    """
    if stopword:
        for token in nltk.word_tokenize(text, preserve_line=preserve_line):
            if token.lower() in STOPWORDS:
                continue
            if lowercase:
                yield token.lower()
            else:
                yield token
    else:
        for token in nltk.word_tokenize(text, preserve_line=preserve_line):
            if lowercase:
                yield token.lower()
            else:
                yield token


def get_token_list(text: str, lowercase: bool = False, stopword: bool = False,
                   preserve_line: bool = True) -> List:
    """
    Get token list from text.
    :param text:
    :param lowercase:
    :param stopword:
    :param preserve_line: keep the preserve the sentence and not sentence tokenize it.
    :return:
    """
    return list(get_tokens(text, lowercase=lowercase, stopword=stopword,
                           preserve_line=preserve_line))


def get_stems(text, lowercase=False, stopword=False, stemmer=DEFAULT_STEMMER) -> Generator:
    """
    Get stems from text.
    N.B.: when stemmer is SnowballStemmer, lowercase is always returned no matter the parameter.
    :param text:
    :param lowercase:
    :param stopword:
    :param stemmer:
    :return:
    """
    for token in get_tokens(text, lowercase=lowercase, stopword=stopword):
        yield stemmer.stem(token)


def get_stem_list(text, lowercase=False, stopword=False, stemmer=DEFAULT_STEMMER) -> List:
    """
    Get stems materialized from text.
    N.B.: when stemmer is SnowballStemmer, lowercase is always returned no matter the parameter.

    :param text:
    :param lowercase:
    :param stopword:
    :param stemmer:
    :return:
    """
    return list(get_stems(text, lowercase=lowercase, stopword=stopword, stemmer=stemmer))


def get_lemmas(text, lowercase=False, stopword=False, lemmatizer=DEFAULT_LEMMATIZER) -> Generator:
    """
    Get lemmas from text.
    :param text:
    :param lowercase:
    :param stopword:
    :param lemmatizer:
    :return:
    """
    tokens = get_token_list(text, lowercase=False, stopword=False)
    pos = nltk.pos_tag(tokens)

    if stopword:
        for i in range(len(tokens)):
            token = pos[i][0]
            wn_pos = get_wordnet_pos(pos[i][1])

            if token.lower() in STOPWORDS:
                continue

            if lowercase:
                yield lemmatizer.lemmatize(token, wn_pos).lower() if wn_pos else lemmatizer.lemmatize(token).lower()
            else:
                yield lemmatizer.lemmatize(token, wn_pos) if wn_pos else lemmatizer.lemmatize(token)
    else:
        for i in range(len(tokens)):
            token = pos[i][0]
            wn_pos = get_wordnet_pos(pos[i][1])

            if lowercase:
                yield lemmatizer.lemmatize(token, wn_pos).lower() if wn_pos else lemmatizer.lemmatize(token).lower()
            else:
                yield lemmatizer.lemmatize(token, wn_pos) if wn_pos else lemmatizer.lemmatize(token)


def get_lemma_list(text, lowercase=False, stopword=False, lemmatizer=DEFAULT_LEMMATIZER) -> List:
    """
    Get lemmas materialized from text.
    """
    return list(get_lemmas(text, lowercase=lowercase, stopword=stopword, lemmatizer=lemmatizer))


def get_verbs(text, lowercase=False, lemmatize=False) -> Generator:
    """
    Get only verbs from text.
    """
    tokens = get_token_list(text)
    pos = nltk.pos_tag(tokens)
    verb_index = [i for i in range(len(pos)) if pos[i][1].startswith("V")]
    if lemmatize:
        lemmas = get_lemma_list(text, lowercase=lowercase)
        for j in verb_index:
            yield lemmas[j]
    else:
        for j in verb_index:
            yield tokens[j].lower() if lowercase else tokens[j]


def get_nouns(text, lowercase=False, lemmatize=False) -> Generator:
    """
    Get only nouns from text.
    """
    tokens = get_token_list(text)
    pos = nltk.pos_tag(tokens)
    noun_index = [i for i in range(len(pos)) if pos[i][1].startswith("N")]
    if lemmatize:
        lemmas = get_lemma_list(text, lowercase=lowercase)
        for j in noun_index:
            yield lemmas[j]
    else:
        for j in noun_index:
            yield tokens[j].lower() if lowercase else tokens[j]


def get_adverbs(text, lowercase=False, lemmatize=False) -> Generator:
    """
    Get only adverbs from text.
    """
    tokens = get_token_list(text)
    pos = nltk.pos_tag(tokens)
    adverb_index = [i for i in range(len(pos)) if pos[i][1].startswith("RB")]
    if lemmatize:
        lemmas = get_lemma_list(text, lowercase=lowercase)
        for j in adverb_index:
            yield lemmas[j]
    else:
        for j in adverb_index:
            yield tokens[j].lower() if lowercase else tokens[j]


def get_adjectives(text, lowercase=False, lemmatize=False) -> Generator:
    """
    Get only adjectives from text.
    """
    tokens = get_token_list(text)
    pos = nltk.pos_tag(tokens)
    adj_index = [i for i in range(len(pos)) if pos[i][1].startswith("JJ")]
    if lemmatize:
        lemmas = get_lemma_list(text, lowercase=lowercase)
        for j in adj_index:
            yield lemmas[j]
    else:
        for j in adj_index:
            yield tokens[j].lower() if lowercase else tokens[j]



import datetime
import regex as re
from typing import List, Tuple

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class YearParser:
    """
    finds years in the string passed
    """
    def __init__(self):
        self.max_year = datetime.datetime.now().year + 1
        self.reg_year = re.compile(r"\d{4}")

    def check_year_ok(self, year: int, min_year: int = 1800, max_year = 0):
        max_year = max_year if max_year > 0 else self.max_year
        return min_year <= year <= max_year

    def get_years_with_coords_from_string(self,
                                          text: str, min_year: int = 1800,
                                          max_year = 0) -> List[Tuple[int, int, int]]:
        years = []  # List[Tuple[int, int, int]]
        for m in self.reg_year.finditer(text):
            year = int(m.group())
            if self.check_year_ok(year, min_year, max_year):
                years.append((year, m.start(), m.end()))
        return years


year_parser = YearParser()



# Imports

from typing import List, Optional

import numpy
import os
import spacy

from lexnlp.extract.ml.classifier.base_token_sequence_classifier_model import BaseTokenSequenceClassifierModel

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


MODULE_PATH = os.path.abspath(os.path.dirname(__file__))
NLP_EN = spacy.load('en_core_web_sm')


# TODO: Refactor to support multiple languages
# TODO: Determine how to introspect model for pos/tag/dep metadata
SPACY_POS_LIST = ['NOUN', 'PROPN', 'PUNCT', 'ADP', 'VERB', 'SPACE', 'DET', 'CCONJ', 'ADJ', 'ADV', 'PART',
                  'NUM', 'X', 'SYM', 'PRON', 'INTJ']
SPACY_TAG_LIST = ['NNP', 'IN', 'NN', 'DT', '""', 'CC', 'NNS', ',', 'JJ', 'VBN', 'VB', '_SP', 'RB', '.', 'CD', 'MD',
                  '-RRB-', '-LRB-', 'VBZ', 'TO', 'VBG', ':', 'POS', 'VBP', "''", '``', 'NNPS', 'WDT', 'VBD', 'XX',
                  'PRP$', 'LS', 'RP', 'HYPH', 'PRP', '$', 'SYM', 'JJR', 'EX', 'NFP', 'RBS', 'RBR', 'WRB', 'JJS', 'WP',
                  'AFX', 'UH']
SPACY_DEP_LIST = ['punct', 'prep', 'pobj', '""', 'det', 'conj', 'cc', 'compound', 'amod', 'ROOT', 'aux', 'nsubj',
                  'dobj', 'advmod', 'appos', 'acl', 'nmod', 'mark', 'nummod', 'auxpass', 'poss', 'advcl', 'case',
                  'nsubjpass', 'agent', 'relcl', 'neg', 'ccomp', 'xcomp', 'acomp', 'attr', 'npadvmod', 'pcomp', 'intj',
                  'prt', 'meta', 'oprd', 'dative', 'parataxis', 'quantmod', 'expl', 'dep', 'csubj', 'preconj']


class SpacyTokenSequenceClassifierModel(BaseTokenSequenceClassifierModel):
    """
    Classifier for generic text sequence objects.
    TODO: Implement skipgram.
    TODO: Implement sum and normalize.
    """

    def __init__(self, letter_set=None, digit_set=None, punc_set=None, symbol_set=None, match_tokens=None,
                 pre_window=0, post_window=0, calculate_sum=False, normalize=False, string_checks=False):
        super().__init__(letter_set=letter_set, digit_set=digit_set, punc_set=punc_set,
                         symbol_set=symbol_set, match_tokens=match_tokens, pre_window=pre_window,
                         post_window=post_window, calculate_sum=calculate_sum,
                         normalize=normalize, string_checks=string_checks)

    def get_feature_list(self,
                         letter_set=None,
                         digit_set=None,
                         punc_set=None,
                         symbol_set=None,
                         pre_window: Optional[int] = None,
                         post_window: Optional[int] = None):
        """
        Return set of features for a character tokenizer.
        """
        # handle params
        if letter_set is None:
            letter_set = self.letter_set

        if digit_set is None:
            digit_set = self.digit_set

        if punc_set is None:
            punc_set = self.punc_set

        if symbol_set is None:
            symbol_set = self.symbol_set

        if pre_window is None:
            pre_window = self.pre_window

        if post_window is None:
            post_window = self.post_window

        # initialize
        token_feature_list = ['position',
                              'length',
                              'mask']

        for pos in SPACY_POS_LIST:
            token_feature_list.append("is_pos_{0}".format(pos))
        for tag in SPACY_TAG_LIST:
            token_feature_list.append("is_tag_{0}".format(tag))
        for dep in SPACY_DEP_LIST:
            token_feature_list.append("is_dep_{0}".format(dep))
        token_feature_list.append("is_pos_other")
        token_feature_list.append("is_tag_other")
        token_feature_list.append("is_dep_other")

        # iterate across window positions
        # pylint: disable=invalid-unary-operand-type
        for i in range(-pre_window, post_window + 1):
            o = str(i)
            token_feature_list.append(o + "_is_start")
            token_feature_list.append(o + "_is_end")
            if self.string_checks:
                token_feature_list.append(o + "_is_title")
                token_feature_list.append(o + "_is_lower")
                token_feature_list.append(o + "_is_upper")

            for c in letter_set:
                token_feature_list.append(o + "_char_" + c)
                token_feature_list.append(o + "_lchar_" + c)
                token_feature_list.append(o + "_first_char_" + c)
                token_feature_list.append(o + "_first_lchar_" + c)
                token_feature_list.append(o + "_last_char_" + c)
                token_feature_list.append(o + "_last_lchar_" + c)

            for c in digit_set:
                token_feature_list.append(o + "_digit_" + c)
                token_feature_list.append(o + "_first_digit_" + c)
                token_feature_list.append(o + "_last_digit_" + c)

            for c in punc_set:
                token_feature_list.append(o + "_punc_" + c)
                token_feature_list.append(o + "_first_punc_" + c)
                token_feature_list.append(o + "_last_punc_" + c)

            for c in symbol_set:
                token_feature_list.append(o + "_symbol_" + c)
                token_feature_list.append(o + "_first_symbol_" + c)
                token_feature_list.append(o + "_last_symbol_" + c)

            token_feature_list.append(o + "_char_other")
            token_feature_list.append(o + "_first_char_other")
            token_feature_list.append(o + "_last_char_other")

            for c in self.unicode_category_set:
                token_feature_list.append(o + "_cat_" + c)
                token_feature_list.append(o + "_first_cat_" + c)
                token_feature_list.append(o + "_last_cat_" + c)
            for c in self.unicode_top_category_set:
                token_feature_list.append(o + "_tcat_" + c)
                token_feature_list.append(o + "_first_tcat_" + c)
                token_feature_list.append(o + "_last_tcat_" + c)

            for token in self.match_tokens:
                token_feature_list.append(o + "_token_" + token)

        if self.calculate_sum:
            token_feature_list.append("sum_char_other")

            for c in self.letter_set:
                token_feature_list.append("sum_char_" + c)

            for c in self.digit_set:
                token_feature_list.append("sum_digit_" + c)

            for c in self.punc_set:
                token_feature_list.append("sum_punc_" + c)

            for c in self.symbol_set:
                token_feature_list.append("sum_symbol_" + c)

        return token_feature_list

    def get_feature_data(self,
                         text: str,
                         feature_mask: List[int] = None):
        """
        Get features based on character model.
        """
        # parse text with spacy
        text_data = [(self.unicode_character_top_category_mapping[
                          c] if c in self.unicode_character_top_category_mapping else "C",
                      self.unicode_character_category_mapping[
                          c] if c in self.unicode_character_category_mapping else "Cc",
                      ord(c)
                      ) for c in text]
        text_lower = text.lower()
        doc = NLP_EN(text)

        # setup return structure
        tokens = []
        num_tokens = len(doc)
        num_features = len(self.feature_list)
        feature_data = numpy.zeros((num_tokens, num_features), dtype=numpy.int8)

        # iterate through tokens
        for i, token in enumerate(doc):
            o = str(0)
            feature_data[i, self._feature_index_map['position']] = i
            feature_data[i, self._feature_index_map['length']] = len(token)
            feature_data[i, self._feature_index_map['mask']] = 0

            if token.pos_ in SPACY_POS_LIST:
                feature_data[i, self._feature_index_map['is_pos_' + token.pos_]] = 1
            else:
                feature_data[i, self._feature_index_map['is_pos_other']] = 1

            if token.tag_ in SPACY_TAG_LIST:
                feature_data[i, self._feature_index_map['is_tag_' + token.tag_]] = 1
            else:
                feature_data[i, self._feature_index_map['is_tag_other']] = 1

            if token.dep_ in SPACY_DEP_LIST:
                feature_data[i, self._feature_index_map['is_dep_' + token.dep_]] = 1
            else:
                feature_data[i, self._feature_index_map['is_dep_other']] = 1

            if self.string_checks:
                feature_data[i, self._feature_index_map[o + "_is_title"]] = token.is_title
                feature_data[i, self._feature_index_map[o + "_is_lower"]] = token.is_lower
                feature_data[i, self._feature_index_map[o + "_is_upper"]] = token.is_upper

            if str(token) in self.match_tokens:
                feature_data[i, self._feature_index_map[o + "_token_" + str(token)]] = 1

            start_pos = token.idx
            end_pos = token.idx + len(token)
            tokens.append((start_pos, end_pos))
            for j in range(start_pos, end_pos):
                c = text[j]
                cl = text_lower[j]
                if feature_mask:
                    feature_data[i, self._feature_index_map['mask']] = max(
                        feature_data[i, self._feature_index_map['mask']], feature_mask[j])

                if c in self.letter_set:
                    feature_data[i, self._feature_index_map[o + "_char_" + c]] += 1
                    feature_data[i, self._feature_index_map[o + "_lchar_" + cl]] += 1
                    if j == start_pos:
                        feature_data[i, self._feature_index_map[o + "_first_char_" + c]] += 1
                        feature_data[i, self._feature_index_map[o + "_first_lchar_" + cl]] += 1
                    if j == end_pos - 1:
                        feature_data[i, self._feature_index_map[o + "_last_char_" + c]] += 1
                        feature_data[i, self._feature_index_map[o + "_last_lchar_" + cl]] += 1
                elif c in self.digit_set:
                    feature_data[i, self._feature_index_map[o + "_digit_" + c]] += 1
                    if j == start_pos:
                        feature_data[i, self._feature_index_map[o + "_first_digit_" + c]] += 1
                    if j == end_pos - 1:
                        feature_data[i, self._feature_index_map[o + "_last_digit_" + c]] += 1
                elif c in self.punc_set:
                    feature_data[i, self._feature_index_map[o + "_punc_" + c]] += 1
                    if j == start_pos:
                        feature_data[i, self._feature_index_map[o + "_first_punc_" + c]] += 1
                    if j == end_pos - 1:
                        feature_data[i, self._feature_index_map[o + "_last_punc_" + c]] += 1
                elif c in self.symbol_set:
                    feature_data[i, self._feature_index_map[o + "_symbol_" + c]] += 1
                    if j == start_pos:
                        feature_data[i, self._feature_index_map[o + "_first_symbol_" + c]] += 1
                    if j == end_pos - 1:
                        feature_data[i, self._feature_index_map[o + "_last_symbol_" + c]] += 1
                else:
                    feature_data[i, self._feature_index_map[o + "_char_other"]] += 1
                    if j == start_pos:
                        feature_data[i, self._feature_index_map[o + "_first_char_other"]] += 1
                    if j == end_pos - 1:
                        feature_data[i, self._feature_index_map[o + "_last_char_other"]] += 1

                feature_data[i, self._feature_index_map[o + "_cat_" + text_data[j][1]]] += 1
                feature_data[i, self._feature_index_map[o + "_tcat_" + text_data[j][0]]] += 1

                if j == start_pos:
                    feature_data[i, self._feature_index_map[o + "_first_cat_" + text_data[j][1]]] += 1
                    feature_data[i, self._feature_index_map[o + "_first_tcat_" + text_data[j][0]]] += 1
                if j == end_pos - 1:
                    feature_data[i, self._feature_index_map[o + "_last_cat_" + text_data[j][1]]] += 1
                    feature_data[i, self._feature_index_map[o + "_last_tcat_" + text_data[j][0]]] += 1

        # handle window feature calculations
        if self.pre_window + self.post_window > 0:
            for i in range(num_tokens):
                j_start = (i - self.pre_window) if (i - self.pre_window >= 0) else 0
                j_end = (i + self.post_window) if (i + self.post_window + 1 < num_tokens) else num_tokens
                for j in range(j_start, j_end):
                    if j == i:
                        continue
                    o = str(j - i)
                    for f in self._base_feature_list:
                        feature_data[i, self._feature_index_map[o + '_' + f]] = \
                            feature_data[j, self._feature_index_map['0_' + f]]

        return feature_data, tokens



from typing import Tuple, List
from lexnlp.extract.common.annotations.text_annotation import TextAnnotation
from lexnlp.utils.map import Map

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class ActAnnotation(TextAnnotation):
    record_type = 'act'
    """
    create an object of ActAnnotation like
    cp = ActAnnotation(name='name', coords=(0, 100), text='text text')
    """
    def __init__(self,
                 coords: Tuple[int, int],
                 locale: str = 'en',
                 act_name: str = '',
                 section: str = '',
                 year: int = None,
                 ambiguous: bool = False,
                 text: str = ''):
        super().__init__(
            name='',
            locale=locale,
            coords=coords,
            text=text)

        self.act_name = act_name
        self.section = section
        self.year = year
        self.ambiguous = ambiguous

    def get_cite_value_parts(self) -> List[str]:
        parts = [self.act_name or '',
                 self.section or '',
                 str(self.year or '')]
        return parts

    def get_dictionary_values(self) -> dict:
        df = Map({
            "tags": {
                'Extracted Entity Name': self.act_name,
                'Extracted Entity Text': self.text
            }
        })
        if self.section:
            df.tags["Extracted Entity Section"] = self.section
        if self.year:
            df.tags["Extracted Entity Year"] = str(self.year)
        if self.ambiguous is not None:
            df.tags["Extracted Entity Ambiguous"] = str(self.ambiguous)
        return df

    def to_dictionary_legacy(self) -> dict:
        return {
            'location_start': self.coords[0],
            'location_end': self.coords[1],
            'act_name': self.act_name,
            'section': self.section,
            'year': str(self.year) if self.year else '',
            'ambiguous': self.ambiguous,
            'value': self.text}



from typing import Tuple, List
from html import escape
from lexnlp.utils.map import Map

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class TextAnnotation:
    """
    It can be a CopyrightAnnotation, CourtAnnotation and so on...

    Example:
        cp = CopyrightAnnotation(name='Siemens', coords=(0, 100), text='text text')
        cp.company = 'Siemens'
        cp.year_start = 1996
        s1 = cp.get_cite()  # '/copyright/Siemens/1996'

        cp.year_end = 2019
        cp.locale = 'en'
        s2 = cp.get_cite()  # '/en/copyright/Siemens/1996/2019'
    """
    record_type = ''

    def __init__(self,
                 name: str,
                 locale: str,
                 coords: Tuple[int, int],
                 text: str = ''):
        self.coords = coords
        self.name = name
        self.text = text
        self.locale = locale

    def __repr__(self):
        s = f'{self.name} [{self.record_type}] at ({self.coords[0]}..{self.coords[1]})'
        if self.locale:
            s += f', loc: {self.locale}'
        return s

    def get_cite(self) -> str:
        path = [self.locale, self.record_type]
        val = self.get_cite_value_parts()
        if val:
            path += val
        return "/" + "/".join([escape(p) for p in path if p])

    def get_cite_value_parts(self) -> List[str]:
        # should be overriden in derived classes
        return [self.name]

    def get_extracted_text(self, full_text: str) -> str:
        # could be overriden
        return full_text[self.coords[0]: self.coords[1]]

    def to_dictionary(self) -> dict:
        df = Map({
            "attrs": {
                "start": self.coords[0],
                "end": self.coords[1]
            },
            "tags": {
                'Extracted Entity Type': self.record_type
            }
        })
        extras = self.get_dictionary_values()
        for key in extras:
            if key not in df:
                df[key] = extras[key]
            else:
                df[key] = {**df[key], **extras[key]}

        return df

    def get_dictionary_values(self) -> dict:
        # could be overriden
        return {}

    @staticmethod
    def safe_cast(val, to_type, default=None):
        try:
            return to_type(val)
        except (ValueError, TypeError):
            return default

    @staticmethod
    def get_int_value(obj, def_value=None) -> int:
        if obj is None:
            return def_value
        if isinstance(obj, int):
            return obj
        try:
            return int(obj)
        # pylint:disable=bare-except
        except:
            return def_value



"""Stanford parsing methods for English.

This module implements an interface to Stanford parsing methods for English, including token and parts of speech.

Todo:
"""

# Imports
import os
from typing import Generator, List

# NLTK imports
from nltk.tag import StanfordPOSTagger
from nltk.tokenize.stanford import StanfordTokenizer

# Project imports
from lexnlp import is_stanford_enabled
from lexnlp.nlp.en.tokens import STOPWORDS, get_lemma_list
from lexnlp.config.stanford import STANFORD_POS_PATH

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


# Setup Stanford POS configuration
try:
    STANFORD_POS_FILE = os.path.join(STANFORD_POS_PATH, "stanford-postagger.jar")
    STANFORD_TOKENIZER = StanfordTokenizer(path_to_jar=STANFORD_POS_FILE)
    STANFORD_DEFAULT_TAG_MODEL = os.path.join(STANFORD_POS_PATH, "models", "english-bidirectional-distsim.tagger")
    STANFORD_TAGGER = StanfordPOSTagger(STANFORD_DEFAULT_TAG_MODEL, STANFORD_POS_FILE)
except LookupError:
    STANFORD_TOKENIZER = STANFORD_TAGGER = None


def check_stanford():
    if not is_stanford_enabled():
        raise RuntimeError("USE_STANFORD is set to False. No Stanford functionality available.")
    if not STANFORD_TOKENIZER:
        raise RuntimeError("USE_STANFORD is set to True."
                           " But default POS tagger jar file is not found."
                           " No Stanford functionality available.")
    if not STANFORD_TAGGER:
        raise RuntimeError("USE_STANFORD is set to True."
                           " But default tagger model file is not found."
                           " No Stanford functionality available.")


def get_tokens_list(text, lowercase=False, stopword=False) -> List:
    """
    Get token list form text using Stanford libraries.
    :param text:
    :param lowercase:
    :param stopword:
    :return:
    """
    return list(get_tokens(text, lowercase=lowercase, stopword=stopword))


def get_tokens(text, lowercase=False, stopword=False) -> Generator:
    """
    Get token generator from text using Stanford libraries.
    :param text:
    :param lowercase:
    :param stopword:
    :return:
    """
    check_stanford()

    if stopword:
        for token in STANFORD_TOKENIZER.tokenize(text):
            if token.lower() in STOPWORDS:
                continue
            if lowercase:
                yield token.lower()
            else:
                yield token
    else:
        for token in STANFORD_TOKENIZER.tokenize(text):
            if lowercase:
                yield token.lower()
            else:
                yield token


def get_verbs(text, lowercase=False, lemmatize=False) -> Generator:
    """
    Get only verbs from text using Stanford libraries.

    :param text:
    :param lowercase:
    :param lemmatize:
    :return:
    """
    check_stanford()

    # Get tokens and tag
    tokens = get_tokens_list(text)
    pos = STANFORD_TAGGER.tag(tokens)

    verb_index = [i for i in range(len(pos)) if pos[i][1].startswith("V")]
    if lemmatize:
        lemmas = get_lemma_list(text, lowercase=lowercase)
        for j in verb_index:
            yield lemmas[j]
    else:
        for j in verb_index:
            yield tokens[j].lower() if lowercase else tokens[j]


def get_nouns(text, lowercase=False, lemmatize=False):
    """
    Get only nouns from text using Stanford libraries.

    :param text:
    :param lowercase:
    :param lemmatize:
    :return:
    """
    check_stanford()

    # Get tokens and tag
    tokens = get_tokens_list(text)
    pos = STANFORD_TAGGER.tag(tokens)

    noun_index = [i for i in range(len(pos)) if pos[i][1].startswith("N")]
    if lemmatize:
        lemmas = get_lemma_list(text, lowercase=lowercase)
        for j in noun_index:
            yield lemmas[j]
    else:
        for j in noun_index:
            yield tokens[j].lower() if lowercase else tokens[j]



# pylint: disable=unused-import

from typing import Pattern, List, Tuple
# pylint: enable=unused-import
import regex as re
from lexnlp.extract.common import year_parser
from lexnlp.extract.common.copyrights.copyright_pattern_found import CopyrightPatternFound
from lexnlp.extract.common.definitions.common_definition_patterns import CommonDefinitionPatterns
from lexnlp.extract.common.pattern_found import PatternFound

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class CopyrightParsingMethods:
    def __init__(self):
        self.trigger_words = ''
        self.reg_trigger_words = None  # type: Pattern
        self.reg_word_c_years = None  # type: List[Pattern]
        self.reg_c_years_word = None  # type: List[Pattern]
        self.init_trigger_words()
        self.init_regexes()

    def init_trigger_words(self):
        # should be overriden
        pass

    def init_regexes(self):
        self.reg_company_name = re.compile(r"[\p{L}\s]+", re.UNICODE | re.IGNORECASE)
        self.reg_trigger_words = re.compile(self.trigger_words,
                                            re.UNICODE | re.IGNORECASE)
        self.reg_word_c_years = re.compile(
            r"[^,;:\|\t]+(%s)\s+\d{4}([–\s-]+\d{4})?" % self.trigger_words,
            re.UNICODE | re.IGNORECASE)
        self.reg_c_years_word = re.compile(
            r"(%s)\s+\d{4}([–\s-]+\d{4})?[^\b;\.]+" % self.trigger_words,
            re.UNICODE | re.IGNORECASE)

    def match_word_c_years(self, phrase: str) -> List[PatternFound]:
        """
        :param phrase: © Siemens 1996 – 2019
        :return: {name: '© Siemens 1996 – 2019', probability: 100, ...}
        """
        if not self.reg_trigger_words.search(phrase):
            return []

        reg = self.reg_word_c_years
        dfs = CommonDefinitionPatterns. \
            collect_regex_matches(phrase, reg, 100,
                                  lambda p, m: m.start(),  # 0 ?
                                  lambda p, m: m.end())  # len(p) ?
        # 'start' - search for company name from the beginning
        patterns = self.pre_process_found_matches(dfs, 'start')
        return patterns

    def match_c_years_word(self, phrase: str) -> List[PatternFound]:
        """
        :param phrase: Copyright 1996 – 2019, Siemens
        :return: {name: '1996 – 2019, Siemens', probability: 100, ...}
        """
        if not self.reg_trigger_words.search(phrase):
            return []

        reg = self.reg_c_years_word
        dfs = CommonDefinitionPatterns. \
            collect_regex_matches(phrase, reg, 100,
                                  lambda p, m: m.start(),
                                  lambda p, m: m.end())  # len(p) ?
        # 'start' - search for company name from the beginning
        patterns = self.pre_process_found_matches(dfs, 'end')
        return patterns

    def pre_process_found_matches(self, matches: List[PatternFound],
                                  company_search_options: str) -> List[CopyrightPatternFound]:
        rst = []
        for match in matches:
            ptrn = CopyrightPatternFound(match)
            # Siemens 1996 – 2019
            # get dates (years)
            years = year_parser.year_parser.get_years_with_coords_from_string(match.name)
            if len(years) == 1:
                ptrn.end_year = years[0][0]
            elif len(years) > 1:
                ptrn.end_year = max(years[0][0], years[1][0])
                ptrn.start_year = min(years[0][0], years[1][0])

            # company name
            company = self.get_company_name_from_match(
                match.name, company_search_options, years)
            if company:
                ptrn.company = company
            rst.append(ptrn)

        return rst

    def get_company_name_from_match(self, text: str,
                                    company_search_options: str,
                                    years: List[Tuple[int, int, int]]) -> str:
        start = 0
        end = -1
        if company_search_options == 'start':
            if len(years) > 0:
                end = years[0][1]
        if company_search_options == 'end':
            if len(years) > 0:
                start = years[-1][2] + 1
        if start == len(text) - 1 or end == 0:
            start = 0
            end = -1

        # try to find company name in source text by 'company_search_options',
        # then, if was not found, find in the whole text
        texts = [text[start: end], text]
        for text in texts:
            words = [w.group().strip(' \t') for w in
                     self.reg_company_name.finditer(text)]
            words = [w for w in words if w]
            phrase = " ".join(words)
            # remove "Copyright" from company name
            phrase = re.sub(self.reg_trigger_words, '', phrase)
            if phrase:
                return phrase

        return ''



import regex as re
from typing import Generator

from lexnlp.extract.common.annotations.percent_annotation import PercentAnnotation
from lexnlp.extract.de.amounts import AmountParserDE

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


amounts_parser = AmountParserDE()
get_amounts = amounts_parser.parse


PERCENT_UNITS_MAP = {
    'prozent': 0.01,
    '%': 0.01
}

PERCENT_PTN = r"""
(?P<text>(?P<num_text>{num_ptn})\s*(?P<unit_name>\w*prozent|%))(?:\W|$)
""".format(num_ptn=amounts_parser.NUM_PTN)
PERCENT_PTN_RE = re.compile(PERCENT_PTN, re.IGNORECASE | re.MULTILINE | re.DOTALL | re.VERBOSE)


def get_percents(text: str, float_digits=4) -> Generator:
    """
    Get percent usages within text.
    :param text:
    :param return_sources:
    :param float_digits:
    :return:
    """
    for ant in get_percent_annotations(text, float_digits):
        yield dict(
                location_start=ant.coords[0],
                location_end=ant.coords[1],
                source_text=ant.text,
                unit_name=ant.sign,
                amount=ant.amount,
                real_amount=ant.fraction)


def get_percent_annotations(text: str, float_digits=4) -> \
        Generator[PercentAnnotation, None, None]:
    """
    Get percent usages within text.
    :param text:
    :param return_sources:
    :param float_digits:
    :return:
    """
    for match in PERCENT_PTN_RE.finditer(text):
        capture = match.capturesdict()
        amount_text = ''.join(capture.get('num_text', ''))
        unit_name = ''.join(capture.get('unit_name', ''))
        amount = list(get_amounts(amount_text, float_digits=float_digits))
        if len(amount) != 1:
            continue
        else:
            amount = amount[0]
        if 'prozent' in unit_name.lower():
            unit_name = 'prozent'
        real_amount = PERCENT_UNITS_MAP.get(unit_name, 0) * amount
        if float_digits:
            real_amount = round(amount, float_digits)
        ant = PercentAnnotation(coords=match.span(),
                                text=''.join(capture.get('text', '')),
                                sign=unit_name,
                                amount=amount,
                                fraction=real_amount,
                                locale='de')
        yield ant


def get_percent_list(*args, **kwargs):
    return list(get_percents(*args, **kwargs))



"""
Configuration for the Stanford NLP library
"""

import os
from lexnlp import get_lib_path

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


# Setup Stanford configuration
STANFORD_VERSION = "2017-06-09"
STANFORD_BASE_PATH = '/usr/lexnlp/libs/stanford_nlp'

STANFORD_POS_PATH = os.path.join(
    STANFORD_BASE_PATH,
    'stanford-postagger-full-{0}'.format(STANFORD_VERSION))
if not os.path.exists(STANFORD_POS_PATH):
    STANFORD_POS_PATH = os.path.join(
        get_lib_path(),
        "stanford_nlp",
        "stanford-postagger-full-{0}".format(STANFORD_VERSION))

STANFORD_NER_PATH = os.path.join(
    STANFORD_BASE_PATH,
    "stanford-ner-{0}".format(STANFORD_VERSION))
if not os.path.exists(STANFORD_NER_PATH):
    STANFORD_NER_PATH = os.path.join(
        get_lib_path(),
        "stanford_nlp",
        "stanford-ner-{0}".format(STANFORD_VERSION))



from abc import abstractmethod
from typing import Optional, Tuple, Union, List, Any, Generator
import string
import num2words
import numpy
import pandas

from lexnlp.extract.ml.classifier.base_token_sequence_classifier_model import BaseTokenSequenceClassifierModel
from lexnlp.extract.ml.detector.detecting_settings import DetectingSettings
from lexnlp.extract.ml.detector.phrase_constructor import PhraseConstructorSettings, PhraseConstructor

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class ArtifactDetector:
    def __init__(self):
        self.model = None  # type: Optional[BaseTokenSequenceClassifierModel]
        self.join_token_settings = PhraseConstructorSettings()

    def load(self, file_path: str):
        self.model = BaseTokenSequenceClassifierModel.load_from_file(file_path)

    def load_compressed(self, file_path: str):
        self.model = BaseTokenSequenceClassifierModel.load_from_file_compressed(file_path)

    def load_from_stream(self, stream: Any):
        self.model = BaseTokenSequenceClassifierModel.load_from_stream(stream)

    def predict(self, sample_df: pandas.DataFrame,
                size_limit: int = 0) -> Tuple[numpy.ndarray, numpy.ndarray]:
        if size_limit:
            sample_df = sample_df.head(size_limit)
        test_feature_data, test_target_data = \
            self.process_sample(sample_df, build_target_data=True, )
        test_predicted = self.model.model.predict(test_feature_data)
        return test_predicted, test_target_data

    @abstractmethod
    def process_sample(self,
                       sample_df: pandas.DataFrame,
                       build_target_data: bool = False) -> Union[numpy.ndarray, Tuple[numpy.ndarray, numpy.ndarray]]:
        raise NotImplementedError('process_sample() should be implemented in derived class')

    def predict_text(self,
                     text: str,
                     join_settings: PhraseConstructorSettings = None,
                     feature_mask: List[int] = None) -> Generator[Tuple[int, int], None, None]:
        feature_data, tokens = self.model.get_feature_data(text, feature_mask)
        predicted_class = self.model.model.predict(feature_data)
        join_settings = join_settings or self.join_token_settings
        yield from PhraseConstructor.join_tokens(
            tokens, predicted_class, settings=join_settings, feature_mask=feature_mask)

    def train_and_save(self,
                       settings: DetectingSettings,
                       train_file: str,
                       train_size: int = -1,
                       save_path: str = '',
                       compress: bool = False) -> None:
        """
        Create a percent identification model using tokens.
        :param settings: Model settings
        :param train_file: File to load training samples from
        :param train_size: Number of records to use
        :param save_path: Output (pickle model) file path
        :param compress: Save compressed file
        """
        # load sample
        train_sample_df = self.read_sample_df(train_file, train_size)

        # setup token match strings in locale
        amount_tokens = self.build_amount_tokens()

        # create model class
        self.train_and_save_on_tokens(amount_tokens, save_path,
                                      settings, train_sample_df,
                                      compress=compress)

    def train_and_save_on_tokens(self,
                                 tokens: List[str],
                                 save_path: str,
                                 settings: DetectingSettings,
                                 train_sample_df: pandas.DataFrame,
                                 punc_set: str = ".,/-",
                                 symbol_set: Optional[str] = None,
                                 string_checks: bool = False,
                                 compress: bool = False):
        self.model = BaseTokenSequenceClassifierModel.get_classifier(
            settings.use_spacy,
            pre_window=settings.pre_window, post_window=settings.post_window,
            match_tokens=tokens, letter_set=string.ascii_letters,
            digit_set=string.digits, punc_set=punc_set, symbol_set=symbol_set,
            string_checks=string_checks)

        # build feature and target training sample
        train_feature_data, train_target_data = self.process_sample(
            train_sample_df, build_target_data=True, )
        # initialize sklearn model based on request
        if settings.model_type == 'extra_trees':
            import sklearn.ensemble

            model = sklearn.ensemble.ExtraTreesClassifier(class_weight="balanced")
        elif settings.model_type == 'random_forest':
            import sklearn.ensemble

            model = sklearn.ensemble.RandomForestClassifier(class_weight="balanced")
        # train
        self.model.train_model(model, train_feature_data, train_target_data)
        if compress:
            self.save_compressed_model(save_path)
        else:
            self.save_model(save_path)

    def save_model(self, save_path: str) -> None:
        self.model.save_in_file(save_path)

    def save_compressed_model(self, save_path):
        self.model.save_in_file_compressed(save_path)

    def read_sample_df(self, train_file: str, train_size: int) -> pandas.DataFrame:
        train_sample_df = pandas.read_csv(train_file, encoding="utf-8", low_memory=False)
        if train_size > 0:
            train_sample_df = train_sample_df.head(train_size)
        return train_sample_df

    def build_amount_tokens(self) -> List[str]:
        amount_tokens = []
        for day in range(1, 101):
            amount_tokens.extend([num2words.num2words(day, to='ordinal'),
                                  num2words.num2words(day, to='ordinal').lower(),
                                  num2words.num2words(day, to='ordinal').upper(),
                                  num2words.num2words(day, to='ordinal_num'),
                                  num2words.num2words(day, to='ordinal_num').lower(),
                                  num2words.num2words(day, to='ordinal_num').upper(),
                                  ])
        amount_tokens.extend(["dozen", "million", "millionth", "billion", "billionth", "trillion", "trillionth"])
        return amount_tokens



from typing import Union, List, Tuple, Generator

import pandas as pd

from lexnlp.extract.common.annotations.geo_annotation import GeoAnnotation
from lexnlp.utils.parse_df import get_entities, get_entity_list, DataframeEntityParser

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


get_geoentities = get_entities


get_geoentity_list = get_entity_list


def get_geoentity_annotations(
                        text: str,
                        config: pd.DataFrame,
                        parse_columns: Union[List[str], Tuple[str]] = None,
                        result_columns: Union[dict, None] = None,
                        preformed_entity: Union[dict, None] = None,
                        priority_sort_column: Union[str, None] = None,
                        priority_sort_ascending: bool = True,
                        cell_values_separator: Union[str, None] = ';',
                        unique_column_values: bool = True) -> Generator[GeoAnnotation, None, None]:
    parser = DeGeoentitiesParser()
    yield from parser.get_geoentity_annotations(text,
                                                config,
                                                parse_columns,
                                                result_columns,
                                                preformed_entity,
                                                priority_sort_column,
                                                priority_sort_ascending,
                                                cell_values_separator,
                                                unique_column_values)


class DeGeoentityColumn:
    """
    dataframe_col_name: column in the source Pandas dataframe
    resulted_name: column in resulted set, produced by DataframeEntityParser
    attr_name: GeoAnnotation property name
    """
    def __init__(self,
                 dataframe_col_name: str,
                 resulted_name: str,
                 attr_name: str):
        self.dataframe_col_name = dataframe_col_name
        self.resulted_name = resulted_name or dataframe_col_name
        self.attr_name = attr_name

    def __repr__(self):
        if self.resulted_name:
            return f'{self.dataframe_col_name} ({self.resulted_name}): {self.attr_name}'
        return f'{self.dataframe_col_name}: {self.attr_name}'


class DeGeoentitiesParser:
    # DF column - GeoAnnotation attribute
    default_annotation_columns = \
        [DeGeoentityColumn('Entity ID', '', 'entity_id'),
         DeGeoentityColumn('Entity Category', '', 'entity_category'),
         DeGeoentityColumn('', 'source', 'source'),
         DeGeoentityColumn('German Name', '', 'name'),
         DeGeoentityColumn('Entity Name', '', 'name_en'),
         DeGeoentityColumn('Entity Priority', '', 'entity_priority'),
         DeGeoentityColumn('ISO-3166-2', '', 'iso_3166_2'),
         DeGeoentityColumn('ISO-3166-3', '', 'iso_3166_3'),
         DeGeoentityColumn('Alias', '', 'alias')]

    default_selecting_columns = ["German Name",
                                 "ISO-3166-2",
                                 "ISO-3166-3",
                                 "Alias"]

    def get_geoentities(self,
                     text: str,
                     config: pd.DataFrame,
                     parse_columns: Union[List[str], Tuple[str]] = None,
                     result_columns: Union[dict, None] = None,
                     preformed_entity: Union[dict, None] = None,
                     priority_sort_column: Union[str, None] = None,
                     priority_sort_ascending: bool = True,
                     cell_values_separator: Union[str, None] = ';',
                     unique_column_values: bool = True) -> Generator:

        parse_columns = parse_columns or self.default_selecting_columns

        yield from DataframeEntityParser(dataframe=config,
                                         parse_columns=parse_columns,
                                         result_columns=result_columns,
                                         preformed_entity=preformed_entity,
                                         priority_sort_column=priority_sort_column,
                                         priority_sort_ascending=priority_sort_ascending,
                                         cell_values_separator=cell_values_separator,
                                         unique_column_values=unique_column_values).get_entities(text)

    def get_geoentity_annotations(
                        self,
                        text: str,
                        config: pd.DataFrame,
                        parse_columns: Union[List[str], Tuple[str]] = None,
                        result_columns: List[DeGeoentityColumn] = None,
                        preformed_entity: Union[dict, None] = None,
                        priority_sort_column: Union[str, None] = None,
                        priority_sort_ascending: bool = True,
                        cell_values_separator: Union[str, None] = ';',
                        unique_column_values: bool = True) -> Generator[GeoAnnotation, None, None]:

        parse_columns = parse_columns or self.default_selecting_columns
        result_columns = result_columns or self.default_annotation_columns
        df_result_columns = {c.dataframe_col_name: c.dataframe_col_name
                             for c in self.default_annotation_columns
                             if c.dataframe_col_name}

        for ent in get_entities(text,
                                config=config,
                                parse_columns=parse_columns,
                                result_columns=df_result_columns,
                                preformed_entity=preformed_entity,
                                priority_sort_column=priority_sort_column,
                                priority_sort_ascending=priority_sort_ascending,
                                cell_values_separator=cell_values_separator,
                                unique_column_values=unique_column_values):
            ant = GeoAnnotation(coords=(ent['location_start'], ent['location_end']),
                                locale='de')
            for col in result_columns:
                setattr(ant, col.attr_name,
                        ent[col.resulted_name or col.dataframe_col_name])
            yield ant



import regex as re
from typing import Generator, List

from lexnlp.extract.common.durations.durations_parser import DurationParser
from lexnlp.extract.common.annotations.duration_annotation import DurationAnnotation
from lexnlp.extract.de.amounts import AmountParserDE

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


amounts_parser = AmountParserDE()
get_amounts = amounts_parser.parse


class DeDurationParser(DurationParser):
    DURATION_MAP = {
        "second": 1 / (60 * 60 * 24),
        "minute": 1 / (60 * 24),
        "hour": 1 / 24,
        "day": 1,
        "week": 7,
        "month": 30,  # 365.25/12.,
        "quarter": 365 / 4,
        "year": 365,  # 365.25,
        "annum": 365,
        "anniversary": 365,
        "anniversaries": 365
    }

    DURATION_TRANSLATION_MAP = {
        "sekunde": 'second',
        "sekunden": 'second',
        "minute": 'minute',
        "minuten": 'minute',
        "stunde": 'hour',
        "stunden": 'hour',
        "tag": 'day',
        "tage": 'day',
        "woche": 'week',
        "wochen": 'week',
        "monat": 'month',
        "monate": 'month',
        "vierteljahr": 'quarter',
        "vierteljahre": 'quarter',
        "jahr": 'year',
        "jahre": 'year',
        "jahres": 'year',
        "jahren": 'year',
    }

    duration_items = sorted(DURATION_TRANSLATION_MAP.keys(), key=len, reverse=True)
    duration_items_joined = '|'.join(duration_items)
    DURATION_MAP_RE = re.compile(duration_items_joined)

    DURATION_PTN = r"""
    (?P<text>(?P<num_text>{num_ptn})?
    (?P<unit_prefix>(?:kalend[ae]r|lebens|actual))?
    (?P<unit_name>{unit_names}))
    (?:\W|$)
    """.format(num_ptn=amounts_parser.NUM_PTN, unit_names=duration_items_joined)

    DURATION_PTN_RE = re.compile(DURATION_PTN, re.IGNORECASE | re.MULTILINE | re.DOTALL | re.VERBOSE)

    INNER_CONJUNCTIONS = ['und', 'plus']

    INNER_PUNCTUATION = re.compile(r'[\s\,]')

    LOCALE = 'de'

    @classmethod
    def get_all_annotations(cls,
                            text: str,
                            float_digits=4) \
            -> List[DurationAnnotation]:

        all_annotations = []
        for match in cls.DURATION_PTN_RE.finditer(text):
            capture = match.capturesdict()
            amount_text = ''.join(capture.get('num_text', ''))
            amounts = list(get_amounts(amount_text, float_digits=float_digits))
            if len(amounts) != 1:
                amount = 1
            else:
                amount = amounts[0]
            unit_name_local = ''.join(capture.get('unit_name', '')).lower()
            unit_prefix = ''.join(capture.get('unit_prefix', '')).lower()
            unit_name_local = cls.DURATION_MAP_RE.findall(unit_name_local)
            if not unit_name_local:
                continue
            unit_name_local = unit_name_local[0]
            unit_name_en = cls.DURATION_TRANSLATION_MAP.get(unit_name_local)

            amount_days = cls.DURATION_MAP[unit_name_en] * amount
            if float_digits:
                amount_days = round(amount_days, float_digits)
            ant = DurationAnnotation(coords=match.span(),
                                     text=''.join(capture.get('text', '')),
                                     amount=amount,
                                     duration_days=amount_days,
                                     duration_type_en=unit_name_en,
                                     duration_type=unit_name_local,
                                     prefix=unit_prefix,
                                     locale=cls.LOCALE)
            all_annotations.append(ant)
        return all_annotations


def get_durations(text: str, float_digits=4) -> Generator:
    for ant in DeDurationParser.get_annotations(text, float_digits):
        yield dict(
                location_start=ant.coords[0],
                location_end=ant.coords[1],
                source_text=ant.text,
                unit_name_local=ant.duration_type,
                unit_name=ant.duration_type_en,
                unit_prefix=ant.prefix,
                amount=ant.amount,
                amount_days=ant.duration_days)


def get_duration_annotations(text: str,
                             float_digits=4) \
        -> Generator[DurationAnnotation, None, None]:
    yield from DeDurationParser.get_annotations(text, float_digits)


def get_duration_annotations_list(text: str,
                                  float_digits=4) \
        -> List[DurationAnnotation]:
    return DeDurationParser.get_annotations(text, float_digits)


def get_duration_list(text: str, float_digits=4):
    return list(get_durations(text, float_digits))



from typing import List, Pattern, Callable

from lexnlp.extract.common.annotations.duration_annotation import DurationAnnotation

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class DurationParser:

    DURATION_MAP = {}

    DURATION_PTN_RE = None  # type:Pattern

    INNER_CONJUNCTIONS = []

    INNER_PUNCTUATION = None  # type:Pattern

    GET_AMOUNTS = None  # type:Callable

    LOCALE = 'en'

    @classmethod
    def get_annotations(cls,
                        text: str,
                        float_digits=4) -> List[DurationAnnotation]:
        all_ants = cls.get_all_annotations(text, float_digits)
        if len(all_ants) < 2:
            return all_ants

        # group annotations
        # like 5 years, 6 months
        # grouped durations are:
        # - bigger timeframe to less timeframe
        # - are separated by punctuation, spaces and conjunctions only
        ant_group = [all_ants[0]]
        all_grouped = [ant_group]

        for a in all_ants[1:]:
            if cls.check_ant_continues_group(ant_group, a, text):
                ant_group.append(a)
            else:
                ant_group = [a]
                all_grouped.append(ant_group)

        # sum group annotations
        annotations = []
        for grp in all_grouped:
            if len(grp) == 1:
                annotations.append(grp[0])
            else:
                summed = cls.sum_annotations(grp)
                annotations.append(summed)
        return annotations

    @classmethod
    def sum_annotations(cls,
                        ant_group: List[DurationAnnotation]) -> \
            DurationAnnotation:
        coords = (ant_group[0].coords[0], ant_group[-1].coords[1])
        rst = DurationAnnotation(coords,
                                 locale=ant_group[0].locale,
                                 is_complex=True)
        rst.duration_days = sum([d.duration_days for d in ant_group])
        rst.amount = rst.duration_days
        rst.duration_type = ant_group[-1].duration_type
        rst.duration_type_en = ant_group[-1].duration_type_en
        return rst

    @classmethod
    def check_ant_continues_group(cls,
                                  ant_group: List[DurationAnnotation],
                                  ant: DurationAnnotation,
                                  text: str) -> bool:
        # the following dur should have shorter timeframe than the preceding one
        a = ant_group[-1]
        b = ant
        adr = cls.DURATION_MAP.get(a.duration_type_en or a.duration_type) or 1
        bdr = cls.DURATION_MAP.get(b.duration_type_en or b.duration_type) or 1
        if bdr >= adr:
            return False

        # the captures should be separated by: spaces, punctuation and
        # conjunctions ("and" in any case)
        intext = text[a.coords[1]:b.coords[0]].lower()
        for conj in cls.INNER_CONJUNCTIONS:
            intext = intext.replace(conj, '')
        intext = cls.INNER_PUNCTUATION.sub('', intext)
        return not intext

    @classmethod
    def get_all_annotations(cls,
                            text: str,
                            float_digits=4) \
            -> List[DurationAnnotation]:
        raise NotImplementedError()



from typing import Union, Tuple

import numpy
import pandas

from lexnlp.extract.ml.detector.detecting_settings import DetectingSettings
from lexnlp.extract.ml.detector.sample_processor import process_sample, get_target_start_end_from_corgetes
from lexnlp.extract.ml.detector.artifact_detector import ArtifactDetector

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class DefinitionPhraseDetector(ArtifactDetector):
    """
    Search for the phrase surrounding the term being defined

    Let the prase be <agrees to serve the Company in such capacity during the
    term of employment (the "Employment Period”).

    ... model_definition will find <term of employment (the "Employment Period”)>
    """

    def process_sample(self,
                       sample_df: pandas.DataFrame,
                       build_target_data: bool = False) -> Union[numpy.ndarray, Tuple[numpy.ndarray, numpy.ndarray]]:
        return process_sample(sample_df, self.model,
                              get_target_start_end=get_target_start_end_from_corgetes,
                              column_name_formatted='labels',
                              feature_mask_column='feature_mask')

    def train_and_save(self,
                       settings: DetectingSettings,
                       train_file: str,
                       train_size: int = -1,
                       save_path: str = '',
                       compress: bool = False) -> None:
        # load sample
        train_sample_df = self.read_sample_df(train_file, train_size)

        self.train_and_save_on_dataframe(settings,
                                         train_sample_df,
                                         save_path,
                                         compress)

    def train_and_save_on_dataframe(
            self,
            settings: DetectingSettings,
            train_sample_df: pandas.DataFrame,
            save_path: str = '',
            compress: bool = False) -> None:
        # setup token match strings in locale
        def_tokens = ['shall', 'includes', 'including',
                      'mean', 'meaning', 'referred', 'known',
                      'refers', 'used', 'hereby', 'interpreted',
                      'defined', 'interpreted', 'collectively',
                      'include', 'individually', 'together',
                      'being', 'foregoing', 'purpose', 'purposed']

        # create model class
        self.train_and_save_on_tokens(def_tokens, save_path,
                                      settings, train_sample_df,
                                      punc_set=":\"()",
                                      compress=compress)



from typing import List

from lexnlp.extract.common.annotations.copyright_annotation import CopyrightAnnotation
from lexnlp.extract.common.annotations.text_annotation import TextAnnotation
from lexnlp.extract.common.pattern_found import PatternFound
from lexnlp.extract.common.text_pattern_collector import TextPatternCollector
from lexnlp.utils.lines_processing.line_processor import LineOrPhrase

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class CopyrightParser(TextPatternCollector):
    def make_annotation_from_pattrn(self,
                                    locale: str,
                                    ptrn: PatternFound,
                                    phrase: LineOrPhrase) -> TextAnnotation:
        ant = CopyrightAnnotation(name=ptrn.name, coords=(ptrn.start, ptrn.end),
                                  text=phrase.text[ptrn.start: ptrn.end],
                                  locale=locale)
        ant.company = ptrn.company  # pattern in in fact CopyrightPatternFound
        ant.year_start = ptrn.start_year
        ant.year_end = ptrn.end_year
        return ant

    def get_annotations_as_dictionaries(self) -> List[dict]:
        dfs = []
        for ant in self.annotations:
            df = ant.to_dictionary()
            dfs.append(df)
        return dfs
